"""Defines several continuous toy problems.

This script defines all the artificial landscapes with
signature [np.ndarray] -> np.ndarray. You might
see that the signs have been flipped from [1] or [2]. This is
because we're dealing with maximizations instead of
minimizations.

[1] Ali R. Al-Roomi (2015). Unconstrained Single-Objective Benchmark
    Functions Repository [https://www.al-roomi.org/benchmarks/unconstrained].
    Halifax, Nova Scotia, Canada: Dalhousie University, Electrical and Computer
    Engineering.
[2] Surjanovic, S. and Bingham, D. Virtual Library of Simulation Experiments:
    Test Functions and Datasets. [https://www.sfu.ca/~ssurjano/optimization.html]
"""

import numpy as np


def ackley_function_01(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    _, d = x.shape

    first = np.exp(-0.2 * np.sqrt((1 / d) * np.sum(x**2, axis=1)))
    second = np.exp((1 / d) * np.sum(np.cos(2 * np.pi * x), axis=1))
    res = 20 * first + second - np.exp(1.0) - 20

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def alpine_01(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    res = -np.sum(np.abs(x * np.sin(x) + 0.1 * x), axis=1)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def alpine_02(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    res = np.prod(np.sin(x) * np.sqrt(x), axis=1)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def bent_cigar(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    first = x[..., 0] ** 2
    second = 1e6 * np.sum(x[..., 1:] ** 1, axis=1)
    res = -(first + second)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def brown(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    first = x[..., :-1] ** 2
    second = x[..., 1:] ** 2

    res = -np.sum(first ** (second + 1) + second ** (first + 1), axis=1)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def chung_reynolds(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    res = -(np.sum(x**2, axis=1) ** 2)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def cosine_mixture(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    first = 0.1 * np.sum(np.cos(5 * np.pi * x), axis=1)
    second = np.sum(x**2, axis=1)

    res = first - second

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def deb_01(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    _, d = x.shape
    res = (1 / d) * np.sum(np.sin(5 * np.pi * x) ** 6, axis=1)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def deb_02(x: np.ndarray) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    _, d = x.shape
    res = (1 / d) * np.sum(np.sin(5 * np.pi * (x ** (3 / 4) - 0.05)) ** 6, axis=1)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def deflected_corrugated_spring(
    x: np.ndarray, alpha: float = 5.0, k: float = 5.0
) -> np.ndarray:
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    sum_of_squares = np.sum((x - alpha) ** 2, axis=1)
    res = -((0.1) * sum_of_squares - np.cos(k * np.sqrt(sum_of_squares)))

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def styblinski_tang(x: np.ndarray, normalize: bool = True) -> np.ndarray:
    """
    This function is maximized at (-2.903534, ..., -2.903534), with
    a value of -39.16599 * d.

    If normalize is True, then the function is normalized to be
    maximized at -39.16599 (i.e. we divide the objective by d).
    """
    assert len(x.shape) == 2
    d = x.shape[1]

    y = (x**4) - (16 * (x**2)) + (5 * x)
    if normalize:
        return -0.5 * np.sum(y, axis=1) / d
    else:
        return -0.5 * np.sum(y, axis=1)


def easom(xy: np.ndarray) -> np.ndarray:
    """
    Easom is very flat, with a maxima at (pi, pi).

    Only works in 2D.
    """
    assert len(xy.shape) == 2, "Easom only works in 2D. "
    assert xy.shape[1] == 2, "Easom only works in 2D. "
    x = xy[..., 0]
    y = xy[..., 1]
    return np.cos(x) * np.cos(y) * np.exp(-((x - np.pi) ** 2 + (y - np.pi) ** 2))


def cross_in_tray(xy: np.ndarray) -> np.ndarray:
    """
    Cross-in-tray has several local maxima in a quilt-like pattern.

    Only works in 2D.
    """
    assert len(xy.shape) == 2, "Easom only works in 2D. "
    assert xy.shape[1] == 2, "Easom only works in 2D. "
    x = xy[..., 0]
    y = xy[..., 1]
    quotient = np.sqrt(x**2 + y**2) / np.pi
    return (
        1 * (np.abs(np.sin(x) * np.sin(y) * np.exp(np.abs(10 - quotient))) + 1) ** 0.1
    )


def egg_holder(xy: np.ndarray) -> np.ndarray:
    """
    The egg holder is especially difficult.

    We only know the optima's location in 2D.
    """
    assert len(xy.shape) == 2, "Easom only works in 2D. "
    assert xy.shape[1] == 2, "Easom only works in 2D. "
    x = xy[..., 0]
    y = xy[..., 1]
    return (y + 47) * np.sin(np.sqrt(np.abs(x / 2 + (y + 47)))) + (
        x * np.sin(np.sqrt(np.abs(x - (y + 47))))
    )


def shifted_sphere(x: np.ndarray) -> np.ndarray:
    """
    The usual squared norm, but shifted away from the origin by a bit.
    Maximized at (1, 1, ..., 1)
    """
    if len(x.shape) == 1:
        # Add a batch dimension if it's missing
        x = x.reshape(-1, x.shape[0])
        batched = False
    else:
        batched = True

    res = -np.sum((x - 1) ** 2, axis=1)

    # Remove the batch dim if it wasn't there in the beginning
    if not batched:
        res = res.squeeze(0)

    return res


def camelback_2d(x: np.ndarray) -> np.ndarray:
    """
    Taken directly from the LineBO repository [1].

    [1] https://github.com/kirschnj/LineBO/blob/master/febo/environment/benchmarks/functions.py
    """
    assert len(x.shape) == 2, "Camelback2D only works in 2D. "
    assert x.shape[1] == 2, "Camelback2D only works in 2D. "
    xx = x[:, 0]
    yy = x[:, 1]
    y = (
        (4.0 - 2.1 * xx**2 + (xx**4) / 3.0) * (xx**2)
        + xx * yy
        + (-4.0 + 4 * (yy**2)) * (yy**2)
    )
    return np.maximum(-y, -2.5)


def hartmann_6d(x: np.ndarray) -> np.ndarray:
    """
    The 6 dimensional Hartmann function.

    Since we are aiming to maximize, we negate the output of the
    function.

    Taken from [2].

    [2] Surjanovic, S. and Bingham, D. Virtual Library of Simulation Experiments:
        Test Functions and Datasets. [https://www.sfu.ca/~ssurjano/optimization.html]
    """
    assert len(x.shape) == 2, "Hartmann6D only works in 6D. "
    assert x.shape[1] == 6, "Hartmann6D only works in 6D. "
    alpha = np.array([1.0, 1.2, 3.0, 3.2])
    A = np.array(
        [
            [10, 3, 17, 3.5, 1.7, 8],
            [0.05, 10, 17, 0.1, 8, 14],
            [3, 3.5, 1.7, 10, 17, 8],
            [17, 8, 0.05, 10, 0.1, 14],
        ]
    )
    P = 10 ** (-4) * np.array(
        [
            [1312, 1696, 5569, 124, 8283, 5886],
            [2329, 4135, 8307, 3736, 1004, 9991],
            [2348, 1451, 3522, 2883, 3047, 6650],
            [4047, 8828, 8732, 5743, 1091, 381],
        ]
    )

    res = []
    # Non-vectorized for now.
    for b in range(x.shape[0]):
        res_at_b = 0
        for i in range(4):
            inner_sum = 0
            for j in range(6):
                inner_sum += A[i][j] * (x[b][j] - P[i][j]) ** 2
            res_at_b += alpha[i] * np.exp(-inner_sum)
        res.append(res_at_b)

    return np.array(res).reshape(-1, 1)


def branin_2d(x: np.ndarray) -> np.ndarray:
    """
    The 2D Branin function.

    Taken from [2]. Notice that we negate the output of the function
    since we are aiming to maximize instead of minimizing.

    [2] Surjanovic, S. and Bingham, D. Virtual Library of Simulation Experiments:
        Test Functions and Datasets. [https://www.sfu.ca/~ssurjano/optimization.html]
    """
    assert len(x.shape) == 2, "Branin2D only works in 2D. "
    assert x.shape[1] == 2, "Branin2D only works in 2D. "
    x1 = x[..., 0]
    x2 = x[..., 1]
    a = 1
    b = 5.1 / (4 * np.pi**2)
    c = 5 / np.pi
    r = 6
    s = 10
    t = 1 / (8 * np.pi)
    y = a * (x2 - b * x1**2 + c * x1 - r) ** 2 + s * (1 - t) * np.cos(x1) + s

    return -y


def rosenbrock(x: np.ndarray, a: float = 1.0, b: float = 100.0):
    """
    Compute the Rosenbrock function.

    Parameters
    ----------
    x (np.ndarray): A 2D numpy array of shape [b, d] where 'b' is the batch size and 'd' is the dimensionality.
    a (float): The 'a' parameter for the Rosenbrock function. Default is 1.
    b (float): The 'b' parameter for the Rosenbrock function. Default is 100.

    Returns
    -------
    np.ndarray: The value of the Rosenbrock function at each point in 'x'. Shape is [b,].
    """
    assert len(x.shape) == 2, "Input x must be a 2D array."
    d = x.shape[1]
    assert d > 1, "Dimensionality must be greater than 1."

    return -(
        np.sum((a - x[:, :-1]) ** 2 + b * (x[:, 1:] - x[:, :-1] ** 2) ** 2, axis=1)
    )


def levy(x: np.ndarray):
    """
    Compute the Levy function.

    Parameters
    ----------
    x (np.ndarray): A 2D numpy array of shape [b, d] where 'b' is the batch size and 'd' is the dimensionality.

    Returns
    -------
    np.ndarray: The value of the Levy function at each point in 'x'. Shape is [b,].

    References
    ----------
    [1] Surjanovic, S. and Bingham, D. Virtual Library of Simulation Experiments:
    Test Functions and Datasets. [https://www.sfu.ca/~ssurjano/optimization.html]
    """
    assert len(x.shape) == 2, "Input x must be a 2D array."
    d = x.shape[1]
    assert d > 0, "Dimensionality must be greater than 0."

    w = 1 + (x - 1) / 4
    term1 = (np.sin(np.pi * w[:, 0])) ** 2
    term3 = (w[:, -1] - 1) ** 2 * (1 + (np.sin(2 * np.pi * w[:, -1])) ** 2)

    wi = w[:, :-1]
    term2 = np.sum((wi - 1) ** 2 * (1 + 10 * (np.sin(np.pi * wi + 1)) ** 2), axis=1)

    return -(term1 + term2 + term3)


if __name__ == "__main__":
    b = branin_2d
    maximal_b = b(np.array([[-np.pi, 12.275], [np.pi, 2.275], [9.42478, 2.475]]))



"""Utilities for verifying the integrity of files downloaded from the internet.

This module contains utilities for checking the integrity of
files downloaded from the internet. This is useful for
reproducibility purposes, and to make sure no malicious
code is being executed.
"""

import hashlib
from pathlib import Path


def compute_md5_from_filepath(filepath: Path, read_mode: str = "rb") -> str:
    """
    Computes the MD5 hex digest from a filepath,
    opening it in binary form by default.

    Parameters
    ----------
    filepath : Path
        The path to the file for which the MD5 digest will be computed.
    read_mode : str, optional
        The mode in which the file will be opened. Default is 'rb' (read in binary mode).

    Returns
    -------
    hex_digest : str
        The MD5 hex digest of the file.

    Examples:
    --------
    >>> compute_md5_from_filepath('/path/to/file.txt')
    'd41d8cd98f00b204e9800998ecf8427e'
    """
    with open(filepath, read_mode) as fp:
        checksum = hashlib.md5(fp.read()).hexdigest()

    return checksum



"""Custom exceptions inside the poli package."""


class PoliException(Exception):
    """Base class for exceptions in the poli package."""

    pass


class BudgetExhaustedException(PoliException):
    """Exception raised when the budget is exhausted."""

    pass


class FoldXNotFoundException(PoliException):
    """Exception raised when FoldX wasn't found in ~/foldx/foldx."""

    pass



import os

import numpy as np
import torch
from torch.utils.data import Dataset

__all__ = [
    "ResidueEnvironment",
    "ResidueEnvironmentsDataset",
    "ToTensor",
    "CavityModel",
    "DownstreamModel",
    "DDGDataset",
    "DDGToTensor",
]


class ResidueEnvironment:
    """
    Residue environment class used to hold necessarry information about the
    atoms of the environment such as atomic coordinates, atom types and the
    class of the missing central amino acid.

    Parameters
    ----------
    xyz_coords: np.ndarray
        Numpy array with shape (n_atoms, 3) containing the x, y, z coordinates.
    atom_types: np.ndarray
        1D numpy array containing the atom types. Integer values in range(6).
    restypes_onehot: np.ndarray
        Numpy array with shape (n_atoms, 21) containing the amino acid
        class of the missing amino acid
    chain_id: str
        Chain id associated to ResidueEnvironment object
    pdb_residue_number: int
        Residue number associated with the ResidueEnvironment object
    pdb_id: str
        PDBID associated with the ResidueEnvironment object
    """

    def __init__(
        self,
        xyz_coords,
        atom_types,
        restype_onehot,
        chain_id,
        pdb_residue_number,
        pdb_id,
    ):
        self.xyz_coords = xyz_coords
        self.atom_types = atom_types
        self.restype_onehot = restype_onehot
        self.restype_index = np.argmax(self.restype_onehot)
        self.chain_id = chain_id
        self.pdb_residue_number = pdb_residue_number
        self.pdb_id = pdb_id

    def __repr__(self):
        return (
            f"<ResidueEnvironment with {self.xyz_coords.shape[0]} atoms. "
            f"pdb_id: {self.pdb_id}, "
            f"chain_id: {self.chain_id}, "
            f"pdb_residue_number: {self.pdb_residue_number}, "
            f"restype_index: {self.restype_index}>"
        )


class ResidueEnvironmentsDataset(Dataset):
    """
    Residue environment dataset class

    Parameters
    ----------
    input_data: Union[List[str], List[ResidueEnvironment]]
        List of parsed pdb filenames in .npz format or list of
        ResidueEnvironment objects
    transform: Callable
        A to-tensor transformer class
    """

    def __init__(
        self,
        input_data,
        transformer,
    ):
        if all(isinstance(x, ResidueEnvironment) for x in input_data):
            self.res_env_objects = input_data
        elif all(isinstance(x, str) for x in input_data):
            self.res_env_objects = self.parse_envs(input_data)
        else:
            raise ValueError(
                "Input data is not of type" "Union[List[str], List[ResidueEnvironment]]"
            )

        self.transformer = transformer

    def __len__(self):
        return len(self.res_env_objects)

    def __getitem__(self, idx):
        sample = self.res_env_objects[idx]
        if self.transformer:
            sample = self.transformer(sample)
        return sample

    def parse_envs(self, npz_filenames):
        res_env_objects = []
        for i in range(len(npz_filenames)):
            coordinate_features = np.load(npz_filenames[i])
            atom_coords_prot_seq = coordinate_features["positions"]
            restypes_onehots_prot_seq = coordinate_features["aa_onehot"]
            selector_prot_seq = coordinate_features["selector"]
            atom_types_flattened = coordinate_features["atom_types_numeric"]

            chain_ids = coordinate_features["chain_ids"]
            pdb_residue_numbers = coordinate_features["residue_numbers"]
            chain_boundary_indices = coordinate_features["chain_boundary_indices"]
            N_residues = selector_prot_seq.shape[0]
            if "bin" in npz_filenames[i]:  # Check if file belongs to Homology data set
                pdb_id = os.path.basename(npz_filenames[i])[:19]
            else:
                pdb_id = os.path.basename(npz_filenames[i])[:4]

            for resi_i in range(N_residues):
                selector = selector_prot_seq[resi_i]
                selector_masked = selector[selector > -1]  # Remove Filler
                coords_mask = (
                    atom_coords_prot_seq[resi_i, :, 0] != -99.0
                )  # Remove filler
                coords = atom_coords_prot_seq[resi_i][coords_mask]
                atom_types = atom_types_flattened[selector_masked]
                restype_onehot = restypes_onehots_prot_seq[resi_i]

                pdb_residue_number = int(pdb_residue_numbers[resi_i])
                # Locate chain id
                for j in range(len(chain_ids)):
                    chain_boundary_0 = chain_boundary_indices[j]
                    chain_boundary_1 = chain_boundary_indices[j + 1]
                    if resi_i in range(chain_boundary_0, chain_boundary_1):
                        chain_id = str(chain_ids[j])
                        break

                res_env_objects.append(
                    ResidueEnvironment(
                        coords,
                        atom_types,
                        restype_onehot,
                        chain_id,
                        pdb_residue_number,
                        pdb_id,
                    )
                )

        return res_env_objects


class ToTensor:
    """
    To-tensor transformer

    Parameters
    ----------
    DEVICE: str
        Either "cuda" (gpu) or "cpu". Is set-able.
    """

    def __init__(self, DEVICE):
        self.device = DEVICE

    def __call__(self, sample):
        """
        Converts single ResidueEnvironment object into x_ and y_
        """

        sample_env = np.hstack(
            [np.reshape(sample.atom_types, [-1, 1]), sample.xyz_coords]
        )

        return {
            "x_": torch.tensor(sample_env, dtype=torch.float32).to(self.device),
            "y_": torch.tensor(np.array(sample.restype_onehot), dtype=torch.float32).to(
                self.device
            ),
        }

    def collate_cat(self, batch):
        """
        Collate method used by the dataloader to collate a
        batch of ResidueEnvironment objects.
        """
        target = torch.cat([torch.unsqueeze(b["y_"], 0) for b in batch], dim=0)

        # To collate the input, we need to add a column which
        # specifies the environtment each atom belongs to
        env_id_batch = []
        for i, b in enumerate(batch):
            n_atoms = b["x_"].shape[0]
            env_id_arr = torch.zeros(n_atoms, dtype=torch.float32).to(self.device) + i
            env_id_batch.append(
                torch.cat([torch.unsqueeze(env_id_arr, 1), b["x_"]], dim=1)
            )
        data = torch.cat(env_id_batch, dim=0)

        return data, target


class CavityModel(torch.nn.Module):
    """
    3D convolutional neural network to missing amino acid classification

    Parameters
    ----------
    DEVICE: str
        Either "cuda" (gpu) or "cpu". Is set-able.
    n_atom_types: int
        Number of atom types. (C, H, N, O, S, P)
    bins_per_angstrom: float
        Number of grid points per Anstrom.
    grid_dim: int
        Grid dimension
    sigma: float
        Standard deviation used for gaussian blurring
    """

    def __init__(
        self,
        get_latent=False,
        n_atom_types=6,
        bins_per_angstrom=1.0,
        grid_dim=18,
        sigma=0.6,
    ):
        super().__init__()

        self.n_atom_types = n_atom_types
        self.get_latent = get_latent
        self.bins_per_angstrom = bins_per_angstrom
        self.grid_dim = grid_dim
        self.sigma = sigma
        self.sigma_p = self.sigma * self.bins_per_angstrom
        self.model()

    def lin_spacing(self):
        lin_spacing = np.linspace(
            start=-self.grid_dim / 2 * self.bins_per_angstrom
            + self.bins_per_angstrom / 2,
            stop=self.grid_dim / 2 * self.bins_per_angstrom
            - self.bins_per_angstrom / 2,
            num=self.grid_dim,
        )
        return lin_spacing

    def model(self):
        """
        TODO: document
        """
        # REMOVED by MGD to squelch a warning
        # from torch, and replaced with the following
        # self.xx, self.yy, self.zz = torch.tensor(
        #     np.meshgrid(
        #         self.lin_spacing(),
        #         self.lin_spacing(),
        #         self.lin_spacing(),
        #         indexing="ij",
        #     ),
        #     dtype=torch.float32,
        # )

        x_, y_, z_ = np.meshgrid(
            self.lin_spacing(),
            self.lin_spacing(),
            self.lin_spacing(),
            indexing="ij",
        )

        self.xx = torch.tensor(x_, dtype=torch.float32)
        self.yy = torch.tensor(y_, dtype=torch.float32)
        self.zz = torch.tensor(z_, dtype=torch.float32)

        self.conv1 = torch.nn.Sequential(
            torch.nn.Conv3d(6, 16, kernel_size=(3, 3, 3), stride=1, padding=1),
            torch.nn.MaxPool3d(
                kernel_size=(2, 2, 2), stride=(2, 2, 2), padding=(0, 0, 0)
            ),
            torch.nn.BatchNorm3d(16),
            torch.nn.LeakyReLU(),
        )
        self.conv2 = torch.nn.Sequential(
            torch.nn.Conv3d(16, 32, kernel_size=(3, 3, 3), stride=1, padding=0),
            torch.nn.MaxPool3d(
                kernel_size=(2, 2, 2), stride=(2, 2, 2), padding=(0, 0, 0)
            ),
            torch.nn.BatchNorm3d(32),
            torch.nn.LeakyReLU(),
        )
        self.conv3 = torch.nn.Sequential(
            torch.nn.Conv3d(32, 64, kernel_size=(3, 3, 3), stride=1, padding=1),
            torch.nn.MaxPool3d(
                kernel_size=(2, 2, 2), stride=(2, 2, 2), padding=(0, 0, 0)
            ),
            torch.nn.BatchNorm3d(64),
            torch.nn.LeakyReLU(),
            torch.nn.Flatten(),
        )
        self.dense1 = torch.nn.Sequential(
            torch.nn.Linear(in_features=64, out_features=128),
            torch.nn.BatchNorm1d(128),
            torch.nn.LeakyReLU(),
        )
        self.dense2 = torch.nn.Linear(in_features=128, out_features=100)
        self.dense3 = torch.nn.Linear(in_features=100, out_features=20)

    def forward(self, x):
        x = self.gaussian_blurring(x)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.dense1(x)
        x = self.dense2(x)
        if self.get_latent == True:
            return x
        elif self.get_latent == False:
            x = torch.nn.functional.leaky_relu(x)
            x = self.dense3(x)
            return x

    def gaussian_blurring(self, x):
        """
        Method that takes 2d torch.Tensor describing the atoms of the batch.

        Parameters
        ----------
        x: torch.Tensor
            Tensor for shape (n_atoms, 5). Each row represents an atom, where:
                column 0 describes the environment of the batch the
                atom belongs to
                column 1 describes the atom type
                column 2,3,4 are the x, y, z coordinates, respectively

        Returns
        -------
        fields_torch: torch.Tensor
            Represents the structural environment with gaussian blurring
            and has shape (-1, self.grid_dim, self.grid_dim, self.grid_dim).
        """
        current_batch_size = torch.unique(x[:, 0]).shape[0]
        fields_torch = torch.zeros(
            (
                current_batch_size,
                self.n_atom_types,
                self.grid_dim,
                self.grid_dim,
                self.grid_dim,
            )
        ).to(x.device)
        for j in range(self.n_atom_types):
            mask_j = x[:, 1] == j
            atom_type_j_data = x[mask_j]
            if atom_type_j_data.shape[0] > 0:
                pos = atom_type_j_data[:, 2:]
                density = torch.exp(
                    -(
                        (torch.reshape(self.xx.to(x.device), [-1, 1]) - pos[:, 0]) ** 2
                        + (torch.reshape(self.yy.to(x.device), [-1, 1]) - pos[:, 1])
                        ** 2
                        + (torch.reshape(self.zz.to(x.device), [-1, 1]) - pos[:, 2])
                        ** 2
                    )
                    / (2 * self.sigma_p**2)
                )

                # Normalize each atom to 1
                density /= torch.sum(density, dim=0)

                # Since column 0 of atom_type_j_data is sorted
                # I can use a trick to detect the boundaries based
                # on the change from one value to another.
                change_mask_j = (
                    atom_type_j_data[:, 0][:-1] != atom_type_j_data[:, 0][1:]
                ).to(atom_type_j_data.device)

                # Add begin and end indices
                ranges_i = torch.cat(
                    [
                        torch.tensor([0]).to(atom_type_j_data.device),
                        torch.arange(atom_type_j_data.shape[0] - 1).to(
                            atom_type_j_data.device
                        )[change_mask_j]
                        + 1,
                        torch.tensor([atom_type_j_data.shape[0]]).to(
                            atom_type_j_data.device
                        ),
                    ]
                )

                # Fill tensor
                for i in range(ranges_i.shape[0]):
                    if i < ranges_i.shape[0] - 1:
                        index_0, index_1 = ranges_i[i], ranges_i[i + 1]
                        fields = torch.reshape(
                            torch.sum(density[:, index_0:index_1], dim=1),
                            [self.grid_dim, self.grid_dim, self.grid_dim],
                        )
                        fields_torch[i, j, :, :, :] = fields
        return fields_torch


class DownstreamModel(torch.nn.Module):
    """
    Downstream FC neural network.
    """

    def __init__(self):
        super().__init__()

        # Model
        self.lin1 = torch.nn.Sequential(
            torch.nn.Linear(142, 128),
            torch.nn.BatchNorm1d(128),
            torch.nn.LeakyReLU(),
        )
        self.lin2 = torch.nn.Sequential(
            torch.nn.Linear(128, 64),
            torch.nn.BatchNorm1d(64),
            torch.nn.LeakyReLU(),
        )
        self.lin3 = torch.nn.Sequential(
            torch.nn.Linear(64, 16),
            torch.nn.BatchNorm1d(16),
            torch.nn.LeakyReLU(),
        )
        self.lin4 = torch.nn.Linear(in_features=16, out_features=1)
        self.sigmoid = torch.nn.Sigmoid()

    # Forward
    def forward(self, x):
        x = self.lin1(x)
        x = self.lin2(x)
        x = self.lin3(x)
        x = self.lin4(x)
        x = self.sigmoid(x)
        return x


class DDGDataset(Dataset):
    """
    ddG dataset
    """

    def __init__(
        self,
        df,
        transformer=None,
    ):
        self.df = df
        self.transformer = transformer

    def __len__(self):
        return self.df.shape[0]

    def __getitem__(self, idx):
        sample = self.df.iloc[idx]
        if self.transformer:
            sample = self.transformer(sample)
        return sample


class DDGToTensor:
    """
    To-tensor transformer for ddG dataframe data

    Parameters
    ----------
    DEVICE: str
        Either "cuda" (gpu) or "cpu". Is set-able.
    """

    def __init__(self, flag, DEVICE):
        self.device = DEVICE
        self.flag = flag

    def __call__(self, sample):
        # Cavity
        atom_types = sample["resenv"].atom_types
        xyz_coords = sample["resenv"].xyz_coords

        x_cavity = np.hstack([np.reshape(atom_types, [-1, 1]), xyz_coords])

        # Downstream
        wt_onehot = np.zeros(20)
        wt_onehot[sample["wt_idx"]] = 1.0
        mt_onehot = np.zeros(20)
        mt_onehot[sample["mt_idx"]] = 1.0
        wt_freq = [sample["wt_nlf"]]
        mt_freq = [sample["mt_nlf"]]
        x_ds = np.hstack([wt_onehot, mt_onehot, wt_freq, mt_freq])
        pdbid = [sample["pdbid"]]
        chainid = [sample["chainid"]]
        variant = [sample["variant"]]

        if self.flag != "pred":
            ddg_fermi = [sample["score_fermi"]]
            ddg = [sample["score"]]
            return {
                "pdbid": pdbid,
                "chainid": chainid,
                "variant": variant,
                "x_cavity": torch.tensor(x_cavity, dtype=torch.float32).to(self.device),
                "x_ds": torch.tensor(x_ds, dtype=torch.float32).to(self.device),
                "score_fermi": torch.tensor(ddg_fermi, dtype=torch.float32).to(
                    self.device
                ),
                "score": torch.tensor(ddg, dtype=torch.float32).to(self.device),
            }
        else:
            return {
                "pdbid": pdbid,
                "chainid": chainid,
                "variant": variant,
                "x_cavity": torch.tensor(x_cavity, dtype=torch.float32).to(self.device),
                "x_ds": torch.tensor(x_ds, dtype=torch.float32).to(self.device),
            }

    def collate_multi(self, batch):
        """
        Collate method used by the dataloader to collate a
        batch of multi data.
        """
        # To collate the input, we need to add a column which
        # specifies the environment each atom belongs to
        env_id_batch = []
        for i, b in enumerate(batch):
            n_atoms = b["x_cavity"].shape[0]
            env_id_arr = torch.zeros(n_atoms, dtype=torch.float32) + i
            env_id_arr = env_id_arr.to(self.device)
            env_id_batch.append(
                torch.cat([torch.unsqueeze(env_id_arr, 1), b["x_cavity"]], dim=1)
            )
        x_cavity_batch = torch.cat(env_id_batch, dim=0)

        # Downstream
        pdbid_batch = [b["pdbid"] for b in batch]
        chainid_batch = [b["chainid"] for b in batch]
        variant_batch = [b["variant"] for b in batch]
        x_ds_batch = torch.cat([torch.unsqueeze(b["x_ds"], 0) for b in batch], dim=0)

        if self.flag != "pred":
            ddg_fermi_batch = torch.cat(
                [torch.unsqueeze(b["score_fermi"], 0) for b in batch], dim=0
            )
            ddg_batch = torch.cat(
                [torch.unsqueeze(b["score"], 0) for b in batch], dim=0
            )
            return (
                pdbid_batch,
                chainid_batch,
                variant_batch,
                x_cavity_batch,
                x_ds_batch,
                ddg_fermi_batch,
                ddg_batch,
            )
        else:
            return pdbid_batch, chainid_batch, variant_batch, x_cavity_batch, x_ds_batch



from __future__ import annotations

from pathlib import Path

from poli.core.problem import Problem
from poli.core.proteins.data_packages import RFPRaspSupervisedDataPackage
from poli.core.util.seeding import seed_python_numpy_and_torch
from poli.objective_repository.rasp.register import RaspBlackBox, RaspProblemFactory


class RFPRaspBlackBox(RaspBlackBox):
    def __init__(
        self,
        additive=True,
        penalize_unfeasible_with=None,
        device=None,
        experiment_id=None,
        tmp_folder=None,
        batch_size=None,
        parallelize=False,
        num_workers=None,
        evaluation_budget=float("inf"),
        force_isolation=False,
    ):
        RFP_PDB_PATH = Path(__file__).parent / "assets"
        wildtype_pdb_path = [
            RFP_PDB_PATH / "2vad_A.pdb",
            RFP_PDB_PATH / "2vae_D.pdb",
            RFP_PDB_PATH / "3e5v_A.pdb",
            RFP_PDB_PATH / "3ned_A.pdb",
            RFP_PDB_PATH / "5lk4_A.pdb",
            RFP_PDB_PATH / "6aa7_A.pdb",
        ]
        chains_to_keep = ["A", "D", "A", "A", "A", "A"]
        super().__init__(
            wildtype_pdb_path,
            additive,
            chains_to_keep,
            penalize_unfeasible_with,
            device,
            experiment_id,
            tmp_folder,
            batch_size,
            parallelize,
            num_workers,
            evaluation_budget,
            force_isolation,
        )


class RFPRaspProblemFactory(RaspProblemFactory):
    def create(
        self,
        additive: bool = True,
        penalize_unfeasible_with: float | None = None,
        device: str | None = None,
        experiment_id: str = None,
        tmp_folder: Path = None,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        rfp_rasp = RFPRaspBlackBox(
            additive=additive,
            penalize_unfeasible_with=penalize_unfeasible_with,
            device=device,
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )
        x0 = rfp_rasp.x0

        data_package = RFPRaspSupervisedDataPackage()

        return Problem(black_box=rfp_rasp, x0=x0, data_package=data_package)



"""
A wrapper around the RFP task of LamBO [1].

References
----------
[1] “Accelerating Bayesian Optimization for Biological Sequence Design with Denoising Autoencoders.”
Stanton, Samuel, Wesley Maddox, Nate Gruver, Phillip Maffettone,
Emily Delaney, Peyton Greenside, and Andrew Gordon Wilson.
arXiv, July 12, 2022. http://arxiv.org/abs/2203.12742.
"""

# AUTHOR: Simon Bartels
# This sequence is problematic due to an X in position 159
# according to PDB this sequence has the most (98%) similarity: SKGEELIKENMHMKLYMEGTVNNHHFKCTTEGEGKPYEGTQTQRIKVVEGGPLPFAFDILATCFMYGSKTFINHTQGIPDFFKQSFPEGFTWERVTTYEDGGVLTVTQDTSLQDGCLIYNVKLRGVNFPSNGPVMQKKTLGWEATTETLYPADGGLEGRCDMALKLVGGGHLHCNLKTTYRSKKPAKNLKMPGVYFVDRRLERIKEADNETYVEQHEVAVARYCDLPSKL
# I've removed the initial SK and some MYG in the middle
# Then it's really just a K in position 159 which is different.
# seq[159] = 'K'  # strings are immutable

PROBLEM_SEQ = "GEELIKENMHMKLYMEGTVNNHHFKCTTEGEGKPYEGTQTQRIKVVEGGPLPFAFDILATCFSKTFINHTQGIPDFFKQSFPEGFTWERVTTYEDGGVLTVTQDTSLQDGCLIYNVKLRGVNFPSNGPVMQKKTLGWEATTETLYPADGGLEGRCDMALXLVGGGHLHCNLKTTYRSXKPAKNLKMPGVYFVDRRLERIKEADNETYVEQHEVAVARYCDLPSKL"
CORRECT_SEQ = "GEELIKENMHMKLYMEGTVNNHHFKCTTEGEGKPYEGTQTQRIKVVEGGPLPFAFDILATCFSKTFINHTQGIPDFFKQSFPEGFTWERVTTYEDGGVLTVTQDTSLQDGCLIYNVKLRGVNFPSNGPVMQKKTLGWEATTETLYPADGGLEGRCDMALKLVGGGHLHCNLKTTYRSKKPAKNLKMPGVYFVDRRLERIKEADNETYVEQHEVAVARYCDLPSKL"

STARTING_N = "n_D0"
ALGORITHM = "ALGORITHM"
BATCH_SIZE = "BATCH_SIZE"

__all__ = [
    "PROBLEM_SEQ",
    "CORRECT_SEQ",
    "STARTING_N",
    "ALGORITHM",
    "BATCH_SIZE",
]



from __future__ import annotations

import logging
from typing import List

import numpy as np
from scipy.stats import genpareto

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.proteins.defaults import AMINO_ACIDS, ENCODING


class RMFIsolatedLogic(AbstractIsolatedFunction):
    """
    RMF internal logic.

    Parameters
    ----------
    wildtype : List[str]
        String sequence of the reference, default: None.
    c : float, optional

    alphabet : List[str]
        Alphabet for the problem, by default AA list provided from poli.core.util.proteins.defaults
    stochasticity: str, optional
    Methods
    -------
    _black_box(x, context=None)
        Main black box method to compute the fitness value of x relative to the WT.

    Raises
    ------
    AssertionError
        If no wildtype sequence is provided.
    """

    def __init__(
        self,
        wildtype: List[str],
        wt_val: float | None = 0.0,
        c: float | None = None,
        kappa: float | None = 0.1,
        alphabet: List[str] | None = None,
        seed: int | None = 0,
    ) -> None:
        """
        Initialize the RMFBlackBox object.
        """
        assert wildtype is not None, (
            "Missing reference input sequence. "
            "Did you forget to pass it to the create of the black box?"
        )
        if not isinstance(wildtype, np.ndarray):
            wildtype = np.array(list(wildtype))
        self.wildtype = wildtype
        self.seed = seed
        if alphabet is None:
            logging.info("using default alphabet AAs.")
            alphabet = AMINO_ACIDS
        assert all(
            [aa in ENCODING.keys() for aa in wildtype]
        ), "Input wildtype elements not in encoding alphabet."
        self.wt_int = np.array([ENCODING.get(aa) for aa in wildtype])
        if c is None:
            c = 1 / (len(alphabet) - 1)
        else:
            c = c
        assert c >= 0, "Invalid c : c > 0 required!"
        logging.info(f"setting c={c}")
        # if c == 0 : uncorrelated HoC landscape (?)
        self.c = c
        self.kappa = kappa
        self.f_0 = (
            wt_val  # in case of standardized observations (around WT) assume w.l.o.g.
        )
        self.alphabet = alphabet
        eta_var = genpareto.stats(c, moments="v")
        self.theta = c / np.sqrt(eta_var)
        self.rng = np.random.default_rng(seed)
        logging.info(f"landscape theta={self.theta}")
        super().__init__()

    @staticmethod
    def f(
        f0: float,
        sigma: np.ndarray,
        sigma_star: np.ndarray,
        c: float,
        kappa: float,
        rand_state,
    ) -> float:
        # from [1] (2) additive term via Hamming distance and constant
        # hamm_dist = hamming(sigma.flatten(), sigma_star.flatten()) # NOTE scipy HD is normalized, DON't USE
        hamm_dist = np.sum(sigma != sigma_star)
        # from [2] nonadd. term is single small value accroding to RV, we use [1]gen.Pareto RV instead of Gaussian
        eta = genpareto.rvs(kappa, size=1, random_state=rand_state)
        # NOTE [1] describes eta as 2^L i.i.d. RV vector, which does not yield a single function value
        f_p = f0 + -c * hamm_dist
        f_val = f_p + eta
        return f_val

    def __call__(self, x: np.ndarray, context=None) -> np.ndarray:
        values = []
        for sequence in x:
            L = len(sequence)
            assert L == self.wildtype.shape[-1], "Inconsistent length: undefined."
            x_int = np.array([ENCODING.get(aa) for aa in sequence])
            val = self.f(
                f0=self.f_0,
                sigma=x_int,
                sigma_star=self.wt_int,
                c=self.c,
                kappa=self.kappa,
                rand_state=self.rng,
            )
            values.append(val)
        return np.array(values).reshape(-1, 1)


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        RMFIsolatedLogic,
        name="rmf_landscape__isolated",
        conda_environment_name="poli__rmf",
        force=True,
    )



"""This test suite contains the core promises we make to the user."""

import numpy as np


def test_creating_an_instance_of_a_black_box():
    from poli.objective_repository import WhiteNoiseBlackBox

    f = WhiteNoiseBlackBox()
    x = np.array([["1", "2", "3"]])
    _ = f(x)


def test_creating_a_problem():
    from poli import create

    white_noise_problem = create(
        name="white_noise",
        seed=42,
        evaluation_budget=100,
    )

    f, x0 = white_noise_problem.black_box, white_noise_problem.x0
    _ = f(x0)

    f.terminate()



"""
Implements a (tunable) fitness landscape of the type RMF [1] for NA [1], AA [2] inputs.

References
----------
[1] Adaptation in Tunably Rugged Fitness Landscapes: The Rough Mount Fuji Model.
    Neidhart J., Szendro I.G., and Krug, J. Genetics 198, 699-721 (2014). https://doi.org/10.1534/genetics.114.167668
[2] Analysis of a local fitness landscape with a model of the rough Mt. Fuji-type landscape: Application to prolyl endopeptidase and thermolysin.
    Aita T., Uchiyama H., et al. Biopolymers 54, 64-79 (2000). https://doi.org/10.1002/(SICI)1097-0282(200007)54:1<64::AID-BIP70>3.0.CO;2-R
"""

from __future__ import annotations

from typing import List

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.core.util.seeding import seed_python_numpy_and_torch


class RMFBlackBox(AbstractBlackBox):
    """
    RMF Black Box implementation.

    Parameters
    ----------
    wildtype : str
        The wildtype amino-acid sequence (aka reference sequence) against which all RMF values are computed against.
    wt_val : float , optional
        The reference value for the WT, zero if observations are standardized, else float value e.g. ddGs
    c : float, optional
        Constant scalar used in RMF computation, by default is the normalizing constant relative to alphabet size
    kappa : float, optional
        Parameterizes the generalized Pareto distribution, by default 0.1 .
        Determines what type of distribution will be sampled from exponential family, Weibull, etc.
    seed : int, optional
        Random seed for replicability of results, by default None.
    alphabet : List[str], optional
        Type of alphabet of the sequences, by default Amino Acids.
        Nucleic Acids possible.
    batch_size : int, optional
        The batch size for parallel evaluation, by default None.
    parallelize : bool, optional
        Flag to parallelize evaluation, by default False.
    num_workers : int, optional
        The number of workers for parallel evaluation, by default None.
    evaluation_budget : int, optional
        The evaluation budget, by default float("inf").
    force_isolation : bool, optional
        Run in an isolated environment and process, by default False.
    """

    def __init__(
        self,
        wildtype: str,
        wt_val: float = 0.0,
        c: float | None = None,
        kappa: float = 0.1,
        seed: int | None = None,
        alphabet: List[str] | None = None,
        batch_size: int | None = None,
        parallelize: bool | None = False,
        num_workers: int | None = None,
        evaluation_budget: int | None = float("inf"),
        force_isolation: bool = False,
    ) -> None:
        """
        Initialize the RMFBlackBox object.

        Parameters
        ----------
        batch_size : int, optional
            The batch-size for parallel evaluation, default: None.
        parallelize : bool, optional
            Flag to parallelize the evaluation, default: False.
        num_workers : int, optional
            Number of workers for parallel evaluation, default: None.
        evaluation_budget : int, optional
            Maximum number of evaluations, default: float("inf").
        force_isolation: bool
            Run the blackbox in an isolated environment, default: False.
        """
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.wildtype = wildtype
        self.wt_val = wt_val
        self.c = c
        self.kappa = kappa
        self.alphabet = alphabet
        self.seed = seed
        self.force_isolation = force_isolation
        _ = get_inner_function(  # NOTE: this implicitly registers
            isolated_function_name="rmf_landscape__isolated",
            class_name="RMFIsolatedLogic",
            module_to_import="poli.objective_repository.rmf_landscape.isolated_function",
            wildtype=self.wildtype,
            wt_val=self.wt_val,
            c=self.c,
            kappa=self.kappa,
            alphabet=self.alphabet,
            seed=self.seed,
            force_isolation=self.force_isolation,
        )

    def _black_box(self, x: np.ndarray, context: None) -> np.ndarray:
        """
        Runs the given input x provided
        in the context with the RMF function and returns the
        total fitness score.

        Parameters
        -----------
        x : np.ndarray
            The input array of strings containing mutations.
        context : None
            The context for the black box computation.

        Returns
        --------
        y: np.ndarray
            The computed fitness score(s) as a numpy array.
        """
        inner_function = get_inner_function(
            isolated_function_name="rmf_landscape__isolated",
            class_name="RMFIsolatedLogic",
            module_to_import="poli.objective_repository.rmf_landscape.isolated_function",
            wildtype=self.wildtype,
            wt_val=self.wt_val,
            c=self.c,
            kappa=self.kappa,
            alphabet=self.alphabet,
            seed=self.seed,
            force_isolation=self.force_isolation,
        )
        return inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="rmf_landscape",
            max_sequence_length=np.inf,
            aligned=True,
            fixed_length=True,
            deterministic=False,
            alphabet=AMINO_ACIDS,  # TODO: differentiate between AA and NA inputs?
            log_transform_recommended=False,
            discrete=True,
        )


class RMFProblemFactory(AbstractProblemFactory):
    """
    Problem factory for the rough Mt Fuji model problem.

    Methods
    -------
    create(...)
        Creates RMF problem instance with specified parameters.
    """

    def create(
        self,
        wildtype: List[str] | str,
        wt_val: float | None = 0.0,
        c: float | None = None,
        kappa: float = 0.1,
        alphabet: List[str] | None = None,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Create a RMFBlackBox object.

        Parameters
        ----------
        wildtype : List[str] | str
            Reference (wild-type) sequence is pseudo-optimum on start.
        wt_val : float, optional
            Reference function value (standardized observations) of WT.
        c : float, optional
            Constant value for function value computation.
            If None passed default value is regularizing 1/(len(alphabet)-1) .
        kappa: float
            Determines generalized Pareto continuous RV.
        alphabet: List[str], optional
            Problem alphabet used, if None is passed default: AMINO_ACIDS.
        seed : int, optional
            Seed for random number generators. If None is passed,
            the seeding doesn't take place.
        batch_size : int, optional
            Number of samples per batch for parallel computation.
        parallelize : bool, optional
            Flag indicating whether to parallelize the computation.
        num_workers : int, optional
            Number of worker processes for parallel computation.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.

        Returns
        -------
        problem : Problem
            A problem instance containing a RMFBlackBox
            function, and initial wildtypes x0.

        Raises
        ------
        ValueError
            If wildtype reference sequence is missing.
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        if wildtype is None:
            raise ValueError("Missing reference sequence!")

        if isinstance(wildtype, str):
            wildtype = list(wildtype)

        f = RMFBlackBox(
            wildtype=wildtype,
            wt_val=wt_val,
            c=c,
            kappa=kappa,
            alphabet=alphabet,
            seed=seed,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )
        x0 = np.array(wildtype).reshape(1, len(wildtype))
        problem = Problem(f, x0)
        return problem



from typing import Literal
from warnings import warn

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.seeding import seed_python_numpy_and_torch
from poli.objective_repository.gfp_cbas.information import AA


class GFPCBasBlackBox(AbstractBlackBox):
    def __init__(
        self,
        problem_type: Literal["gp", "vae", "elbo"],
        functional_only: bool = False,
        ignore_stops: bool = True,
        unique=True,
        n_starting_points: int = 1,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        seed: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
        negate: bool = False,
    ):
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.problem_type = problem_type
        self.functional_only = functional_only
        self.ignore_stops = ignore_stops
        self.unique = unique
        self.force_isolation = force_isolation
        self.n_starting_points = n_starting_points
        self.seed = seed
        self.negate = negate

        inner_function = get_inner_function(
            isolated_function_name="gfp_cbas__isolated",
            class_name="GFPCBasIsolatedLogic",
            module_to_import="poli.objective_repository.gfp_cbas.isolated_function",
            seed=self.seed,
            force_isolation=self.force_isolation,
            quiet=False,
            problem_type=self.problem_type,
            info=self.get_black_box_info(),
            n_starting_points=self.n_starting_points,
            functional_only=self.functional_only,
            ignore_stops=self.ignore_stops,
            unique=self.unique,
        )
        self.x0 = inner_function.x0

    def _black_box(self, x: np.array, context=None) -> np.ndarray:
        """
        x is encoded sequence return function value given problem name
        """
        inner_function = get_inner_function(
            isolated_function_name="gfp_cbas__isolated",
            class_name="GFPCBasIsolatedLogic",
            module_to_import="poli.objective_repository.gfp_cbas.isolated_function",
            seed=self.seed,
            force_isolation=self.force_isolation,
            quiet=True,
            problem_type=self.problem_type,
            info=self.get_black_box_info(),
            n_starting_points=self.n_starting_points,
            functional_only=self.functional_only,
            ignore_stops=self.ignore_stops,
            unique=self.unique,
        )
        if self.negate:
            return -inner_function(x, context=context)
        return inner_function(x, context=context)

    def __iter__(self, *args, **kwargs):
        warn(f"{self.__class__.__name__} iteration invoked. Not implemented!")

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="gfp_cbas",
            max_sequence_length=237,  # max len of aaSequence
            aligned=True,
            fixed_length=True,
            deterministic=False,
            alphabet=AA,
            log_transform_recommended=False,
            discrete=True,
            fidelity=None,
            padding_token="",
        )


class GFPCBasProblemFactory(AbstractProblemFactory):
    def __init__(self, problem_type: str = "gp") -> None:
        super().__init__()
        if problem_type.lower() not in ["gp", "vae", "elbo"]:
            raise NotImplementedError(
                f"Specified problem type: {problem_type} does not exist!"
            )
        self.problem_type = problem_type.lower()

    def create(
        self,
        problem_type: Literal["gp", "vae", "elbo"] = "gp",
        n_starting_points: int = 1,
        functional_only: bool = False,
        unique: bool = True,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        negate: bool = False,
    ) -> Problem:
        """
        Seed value required to shuffle the data, otherwise CSV asset data index unchanged.
        We optimize with respect to one GFP WT sequence by default.
        If more starting points are requested the sequences are provided at random.
        """
        self.problem_type = problem_type
        if seed is not None:
            seed_python_numpy_and_torch(seed)
        problem_info = self.get_problem_name()
        f = GFPCBasBlackBox(
            problem_type=problem_type,
            functional_only=functional_only,
            unique=unique,
            n_starting_points=n_starting_points,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            seed=seed,
            evaluation_budget=evaluation_budget,
            negate=negate,
        )
        x0 = f.x0

        problem = Problem(f, x0)

        return problem



# plt.rcParams["figure.dpi"] = 300
# plt.rcParams["figure.figsize"] = [8.0, 8.0]
# plt.rcParams.update({"font.size": 14})
import glob
import math
import os
import pickle

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# import mpl_scatter_density
import ptitprince as pt
import seaborn as sns
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.offsetbox import AnchoredText
from matplotlib.ticker import FormatStrFormatter
from scipy import stats
from sklearn.metrics import mean_absolute_error

white_viridis = LinearSegmentedColormap.from_list(
    "white_viridis",
    [
        (0, "#ffffff"),
        (1e-20, "#440053"),
        (0.2, "#404388"),
        (0.4, "#2a788e"),
        (0.6, "#21a784"),
        (0.8, "#78d151"),
        (1, "#fde624"),
    ],
    N=256,
)


def learning_curve_cavity(acc_val_list, acc_train_list, loss_train_list):
    """
    Plots the cavity model learning curve
    """
    fig, ax1 = plt.subplots()
    epochs = np.arange(len(acc_val_list)) + 1
    lns1 = ax1.plot(
        epochs, loss_train_list, label="Train: Cross-entropy loss", color="blue"
    )
    ax1.set_ylabel("Cross-entropy loss")
    ax1.set_xticks([1, 3, 5, 7, 9, 11, 13, 15, 17, 19])
    ax1.set_xlabel("Training epochs")
    ax2 = ax1.twinx()
    lns2 = ax2.plot(epochs, acc_train_list, label="Train: Accuracy", color="green")
    lns3 = ax2.plot(epochs, acc_val_list, label="Val: Accuracy", color="orange")
    ax2.set_ylabel("Accuracy")
    lns = lns1 + lns2 + lns3
    labs = [l.get_label() for l in lns]
    ax2.legend(lns, labs, loc="center right")
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/cavity_models/cavity_learning_curve.pdf",
        bbox_inches="tight",
    )
    plt.close()


def learning_curve_ds(pearson_r_list, val_loss_list, train_loss_list, model_idx):
    """
    Plots the downstream model learning curve
    """
    fig, ax1 = plt.subplots()
    epochs = np.arange(len(pearson_r_list)) + 1
    lns1 = ax1.plot(
        epochs, pearson_r_list, label="Val: Pearson $\\rho$", color="orange"
    )
    ax1.set_ylabel("Pearson $\\rho$")
    ax1.set_xlabel("Training epochs")
    ax1.set_ylim(0.70, 0.80)
    ax1.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))
    ax2 = ax1.twinx()
    lns2 = ax2.plot(epochs, val_loss_list, label="Val: MAE$_{F}$ loss", color="green")
    lns3 = ax2.plot(
        epochs, train_loss_list, label="Train: MAE$_{F}$ loss", color="blue"
    )
    ax2.set_ylabel("MAE$_{F}$ loss")
    ax2.yaxis.set_major_formatter(FormatStrFormatter("%.3f"))
    lns = lns1 + lns2 + lns3
    labs = [l.get_label() for l in lns]
    ax2.legend(lns, labs, loc="center right")
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/ds_models/ds_model_{model_idx}/learning_curve.pdf",
        bbox_inches="tight",
    )
    plt.close()


def learning_curve_ds_with_errorbars():
    """
    Plots the downstream model learning curve with error bars
    """

    model_pearson_all = np.zeros((10, 20))
    model_val_loss_all = np.zeros((10, 20))
    model_train_loss_all = np.zeros((10, 20))

    model_files = glob.glob(f"{os.path.dirname(os.getcwd())}/output/ds_models/**/*.pkl")
    for i, model_file in enumerate(model_files):
        with open(f"{model_file}", "rb") as handle:
            e = pickle.load(handle)
            model_pearson_all[i, :] = e[0]
            model_val_loss_all[i, :] = e[1]
            model_train_loss_all[i, :] = e[2]

    fig, ax1 = plt.subplots()
    epochs = np.arange(model_train_loss_all.shape[1]) + 1
    lns1 = ax1.errorbar(
        epochs,
        np.mean(model_pearson_all, axis=0),
        yerr=np.std(model_pearson_all, axis=0),
        label="Val: Pearson $\\rho$",
        color="orange",
    )
    ax1.set_ylabel("Pearson $\\rho$")
    ax1.set_xlabel("Training epochs")
    ax1.set_ylim(0.70, 0.80)
    ax1.xaxis.set_major_formatter(FormatStrFormatter("%.0f"))
    ax1.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))
    ax2 = ax1.twinx()
    lns2 = ax2.errorbar(
        epochs,
        np.mean(model_val_loss_all, axis=0),
        yerr=np.std(model_val_loss_all, axis=0),
        label="Val: MAE$_{F}$ loss",
        color="green",
    )
    lns3 = ax2.errorbar(
        epochs,
        np.mean(model_train_loss_all, axis=0),
        yerr=np.std(model_train_loss_all, axis=0),
        label="Train: MAE$_{F}$ loss",
        color="blue",
    )
    ax2.set_ylabel("MAE$_{F}$ loss")
    ax2.yaxis.set_major_formatter(FormatStrFormatter("%.3f"))
    ax2.set_ylim(0.050, 0.110)
    lns = [lns1] + [lns2] + [lns3]
    labs = [l.get_label() for l in lns]
    ax2.legend(lns, labs, loc="center right")
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/ds_models/ds_learning_curve_with_errorbars.pdf",
        bbox_inches="tight",
    )
    plt.close()


def hist_plot(df, dataset_key):
    """
    Plots the histogram of ddGs from a dataframe
    """

    # Plot all data
    x = df["score_ml"]
    fig, ax = plt.subplots()
    ax.hist(x)
    ax.set_xlabel("ML \u0394\u0394G")
    ax.set_ylabel("Count")
    ax.set_title(f"{dataset_key} - All")
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/total_pred_hist.pdf",
        bbox_inches="tight",
    )
    plt.close()

    # Plot individual proteins
    prot_names = df["pdbid"].unique()
    for pdbid in prot_names:
        # Extract protein data
        df_pdbid = df[df["pdbid"] == pdbid]

        # Plot
        x = df_pdbid["score_ml"]
        fig, ax = plt.subplots()
        ax.hist(x)
        ax.set_xlabel("ML \u0394\u0394G")
        ax.set_ylabel("Count")
        ax.set_title(f"{dataset_key} - {pdbid}")
        fig.savefig(
            f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/{pdbid}_pred_hist.pdf",
            bbox_inches="tight",
        )
        plt.close()


def hist_plot_all(df, dataset_key):
    """
    Plots the histogram of ddGs from a dataframe.
    This function is specific to the large cytosolic data set.
    """

    # Plot all data
    x = df["score_ml"]
    print(len(x))
    fig, ax = plt.subplots()
    ax.hist(x, color="blue", bins=100, weights=[1 / 10**3] * len(x))
    ax.set_xlabel("RaSP \u0394\u0394G [kcal/mol]")
    ax.set_ylabel("Variant count x $10^3$")
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/total_pred_hist.pdf",
        bbox_inches="tight",
    )
    plt.close()


def make_scatter(x, y, x_vs_y, dataset_key, pdb):
    """
    Plots scatter a scatter plot from a ddG dataframe
    """

    # Compute statistics
    pearson_r = stats.pearsonr(x, y)[0]
    mae = None
    if dataset_key != "VAMP" and (x_vs_y != "exp_vs_ml" or x_vs_y != "exp_vs_ml"):
        mae = mean_absolute_error(x, y)

    # Make plot
    fig = plt.figure()

    # Add data points
    if len(x) > 400:
        ax = fig.add_subplot(1, 1, 1, projection="scatter_density")
        density = ax.scatter_density(x, y, cmap=white_viridis)
        fig.colorbar(density, label="Number of points per pixel")
    else:
        ax = fig.add_subplot(1, 1, 1)
        ax.scatter(x, y, s=2, c="darkblue")

    # Set xy limits, labels and title
    if dataset_key == "Rosetta":
        xylim = ((-1, 7), (-1, 7))
    elif (
        dataset_key == "Protein_G"
        or dataset_key == "ProTherm"
        or "ProTherm_with_Protein_G"
    ):
        xylim = ((-2, 7), (-2, 7))
    elif dataset_key == "VAMP":
        xylim = ((-0.5, 2), (-2, 7))
    ax.set_xlim(xylim[0][0], xylim[0][1])
    ax.set_ylim(xylim[1][0], xylim[1][1])
    if x_vs_y == "exp_vs_ml":
        ax.set_xlabel("Experimental \u0394\u0394G [kcal/mol]")
        if x_vs_y == "exp_vs_ml":
            ax.set_xlabel("VAMP-seq score [a.u.]")
        ax.set_ylabel("RaSP \u0394\u0394G [kcal/mol]")
    elif x_vs_y == "exp_vs_rosetta":
        ax.set_xlabel("Experimental \u0394\u0394G [kcal/mol]")
        if x_vs_y == "exp_vs_ml":
            ax.set_xlabel("VAMP-seq score [a.u.]")
        ax.set_ylabel("Rosetta \u0394\u0394G [kcal/mol]")
    elif x_vs_y == "rosetta_vs_ml":
        ax.set_xlabel("Rosetta \u0394\u0394G [kcal/mol]")
        ax.set_ylabel("RaSP \u0394\u0394G [kcal/mol]")
    ax.xaxis.set_major_formatter(FormatStrFormatter("%.1f"))
    ax.yaxis.set_major_formatter(FormatStrFormatter("%.1f"))

    # Add textbox
    text = f"Pearson $\\rho$: {pearson_r:.2f}"
    if mae != None:
        text += f"\nMAE: {mae:.2f}"
    anchored_text = AnchoredText(text, loc="upper left")
    ax.add_artist(anchored_text)

    # Save
    if pdb == None:
        plot_name = "total"
    else:
        plot_name = pdb
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/{plot_name}_{x_vs_y}_scatter.pdf",
        bbox_inches="tight",
    )
    plt.close()


def make_scatter_combined(df, dataset_key, x_vs_y, prot_names):
    """
    Plots a scatter plot from a ddG dataframe with multiple proteins
    """

    if dataset_key == "ProTherm_with_Protein_G":
        # Plot each protein
        fig = plt.figure()
        colors = ["red", "fuchsia", "blue", "green", "indigo"]

        for i, pdbid in enumerate(prot_names):
            # Select protein data
            df_pdbid = df[df["pdbid"] == pdbid]
            if x_vs_y == "exp_vs_ml":
                x = df_pdbid["score_exp"]
                y = df_pdbid["score_ml"]
            elif x_vs_y == "exp_vs_rosetta":
                x = df_pdbid["score_exp"]
                y = df_pdbid["score_rosetta"]

            y = y[(x >= -1) & (x <= 7)]
            x = x[(x >= -1) & (x <= 7)]

            # Compute statistics
            pearson_r = stats.pearsonr(x, y)[0]

            # Make plot
            ax = fig.add_subplot(1, 1, 1)
            ax.xaxis.set_major_formatter(FormatStrFormatter("%.1f"))
            ax.yaxis.set_major_formatter(FormatStrFormatter("%.1f"))

            # Add data points
            ax = fig.add_subplot(1, 1, 1)
            ax.scatter(
                x,
                y,
                s=2,
                c=colors[i],
                label=f"{prot_names[i]}, Pearson $\\rho$: {pearson_r:.2f}",
            )

            # Set xy limits, labels and title
            xylim = ((-2, 7), (-2, 7))
            ax.set_xlim(xylim[0][0], xylim[0][1])
            ax.set_ylim(xylim[1][0], xylim[1][1])

        if x_vs_y == "exp_vs_ml":
            ax.set_xlabel("Experimental \u0394\u0394G [kcal/mol]")
            ax.set_ylabel("RaSP \u0394\u0394G [kcal/mol]")
        if x_vs_y == "exp_vs_rosetta":
            ax.set_xlabel("Experimental \u0394\u0394G [kcal/mol]")
            ax.set_ylabel("Rosetta \u0394\u0394G [kcal/mol]")

        # Add legends
        plt.legend(loc="upper left")

        # Save
        plot_name = "combined"
        fig.savefig(
            f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/{dataset_key}_{plot_name}_{x_vs_y}_scatter.pdf",
            bbox_inches="tight",
        )
        plt.close()

    if dataset_key == "VAMP":
        # Map PDB ids to protein names
        name_dict = {
            "5BON": "NUDT15 (PDB: 5BON)",
            "2H11": "TPMT (PDB: 2H11)",
            "1D5R": "PTEN (PDB: 1D5R)",
        }

        # Plot each protein
        for i, pdbid in enumerate(prot_names):
            # Select protein data
            df_pdbid = df[df["pdbid"] == pdbid]
            x = df_pdbid["score_exp"]
            y_ml = df_pdbid["score_ml"]
            y_ml = y_ml[(x >= -1) & (x <= 7)]
            y_rosetta = df_pdbid["score_rosetta"]
            y_rosetta = y_rosetta[(x >= -1) & (x <= 7)]
            x = x[(x >= -1) & (x <= 7)]

            # Compute statistics
            pearson_r_ml = stats.pearsonr(x, y_ml)[0]
            pearson_r_rosetta = stats.pearsonr(x, y_rosetta)[0]

            # Make plot
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.xaxis.set_major_formatter(FormatStrFormatter("%.1f"))
            ax.yaxis.set_major_formatter(FormatStrFormatter("%.1f"))

            # Add data points
            ax.scatter(
                x,
                y_ml,
                s=2,
                c="red",
                label=f"RaSP - Pearson $\\rho$: {pearson_r_ml:.2f}",
            )
            ax.scatter(
                x,
                y_rosetta,
                s=2,
                c="blue",
                label=f"Rosetta - Pearson $\\rho$: {pearson_r_rosetta:.2f}",
            )

            # Set xy limits, labels and title
            xylim = ((-0.5, 2), (-1, 7))
            ax.set_xlim(xylim[0][0], xylim[0][1])
            ax.set_ylim(xylim[1][0], xylim[1][1])

            # Set title
            ax.set_title(name_dict[pdbid])

            # if x_vs_y == "exp_vs_ml":
            #    ax.set_xlabel("VAMP-seq score [a.u.]")
            #    ax.set_ylabel("RaSP \u0394\u0394G [kcal/mol]")
            ax.set_xlabel("VAMP-seq score [a.u.]")
            ax.set_ylabel("\u0394\u0394G [kcal/mol]")

            # Add legends
            legend = plt.legend(loc="upper left")

            # Make the font larger for this particular plot
            for item in (
                [ax.title, ax.xaxis.label, ax.yaxis.label]
                + ax.get_xticklabels()
                + ax.get_yticklabels()
                + legend.get_texts()
            ):
                item.set_fontsize(22)

            # Save
            plot_name = "combined"
            fig.savefig(
                f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/{dataset_key}_{plot_name}_{x_vs_y}_scatter_{pdbid}.pdf",
                bbox_inches="tight",
            )
            plt.close()


def scatter_plots(df, dataset_key):
    """
    Function for plotting multiple scatter plots depending on the dataset
    """

    # Create plot: ML vs Rosetta - All
    x = df["score_rosetta"]
    y = df["score_ml"]
    y = y[(x >= -1) & (x <= 7)]
    x = x[(x >= -1) & (x <= 7)]
    make_scatter(x, y, "rosetta_vs_ml", dataset_key, None)

    # Create plot: ML vs Rosetta - Individual
    prot_names = df["pdbid"].unique()
    for pdbid in prot_names:
        # Extract protein data
        df_pdbid = df[df["pdbid"] == pdbid]
        x = df_pdbid["score_rosetta"]
        y = df_pdbid["score_ml"]
        y = y[(x >= -1) & (x <= 7)]
        x = x[(x >= -1) & (x <= 7)]
        make_scatter(x, y, "rosetta_vs_ml", dataset_key, pdbid)

    if dataset_key != "Rosetta" and dataset_key != "Xstal_vs_AF2":
        # Create plot: ML vs Exp - All
        x = df["score_exp"]
        y = df["score_ml"]
        make_scatter(x, y, "exp_vs_ml", dataset_key, None)

        # Create plot: ML vs Exp - Individual
        prot_names = df["pdbid"].unique()
        for pdbid in prot_names:
            # Extract protein data
            df_pdbid = df[df["pdbid"] == pdbid]
            x = df_pdbid["score_exp"]
            y = df_pdbid["score_ml"]
            make_scatter(x, y, "exp_vs_ml", dataset_key, pdbid)

        # Create plot: Rosetta vs Exp - All
        x = df["score_exp"]
        y = df["score_rosetta"]
        make_scatter(x, y, "exp_vs_rosetta", dataset_key, None)

        # Create plot: Rosetta vs Exp - Individual
        prot_names = df["pdbid"].unique()
        for pdbid in prot_names:
            # Extract protein data
            df_pdbid = df[df["pdbid"] == pdbid]
            x = df_pdbid["score_exp"]
            y = df_pdbid["score_rosetta"]
            make_scatter(x, y, "exp_vs_rosetta", dataset_key, pdbid)

    if dataset_key == "ProTherm_with_Protein_G":
        # Create plot: ML vs Exp - All --> Individual colors
        prot_names = df["pdbid"].unique()
        make_scatter_combined(df, dataset_key, "exp_vs_ml", prot_names)
        make_scatter_combined(df, dataset_key, "exp_vs_rosetta", prot_names)

    if dataset_key == "VAMP":
        # Create plot: ML vs VAMP & Rosetta vs VAMP --> Individual colors
        prot_names = df["pdbid"].unique()
        make_scatter_combined(df, dataset_key, "exp_vs_ml", prot_names)


def loss_analysis(df, dataset_key):
    """
    Plots three 20x20 matrices:
       1) Loss per amino acid
       2) ddG variance per amino acid
       3) Residue count in high-loss region per amino acid
    """

    # Order aa by chemistry (aromatic, hydrophobic, hydroxylic, polar, negatively charged)
    aa_list = [
        "W",
        "Y",
        "F",
        "L",
        "M",
        "I",
        "V",
        "G",
        "A",
        "S",
        "T",
        "C",
        "N",
        "H",
        "Q",
        "K",
        "R",
        "D",
        "E",
        "P",
    ]

    # Extract wt and mt
    df["wt"] = df["variant"].str.extract("(^[a-zA-Z])")
    df["mt"] = df["variant"].str.extract("([a-zA-Z]$)")

    # Compute MAEs
    df["mae_ml_rosetta"] = (df["score_ml"] - df["score_rosetta"]).abs()
    if dataset_key != "Rosetta" and dataset_key != "Xstal_vs_AF2":
        df["mae_ml_exp"] = (df["score_ml"] - df["score_exp"]).abs()

    # Divide into high- and low-loss
    df_highloss_rosetta = df[df["mae_ml_rosetta"] > 2]
    df_lowloss_rosetta = df[df["mae_ml_rosetta"] < 2]
    if dataset_key != "Rosetta":
        df_highloss_exp = df[df["mae_ml_rosetta"] > 2]
        df_highloss_exp = df[df["mae_ml_rosetta"] < 2]

    # Compute statistics per mutation
    mat_loss_ml_rosetta = np.zeros((20, 20))
    mat_var_rosetta = np.zeros((20, 20))
    mat_highloss_count_rosetta = np.zeros((20, 20))
    if dataset_key != "Rosetta":
        mat_loss_ml_exp = np.zeros((20, 20))
        mat_var_exp = np.zeros((20, 20))
        mat_highloss_count_exp = np.zeros((20, 20))

    for i, res_wt in enumerate(aa_list):
        for j, res_mt in enumerate(aa_list):
            mat_loss_ml_rosetta[i, j] = df["mae_ml_rosetta"][
                (df["wt"] == res_wt) & (df["mt"] == res_mt)
            ].mean()
            mat_var_rosetta[i, j] = df["score_rosetta_fermi"][
                (df["wt"] == res_wt) & (df["mt"] == res_mt)
            ].var()
            if len(df[(df["wt"] == res_wt) & (df["mt"] == res_mt)]) > 0:
                mat_highloss_count_rosetta[i, j] = len(
                    df_highloss_rosetta[
                        (df_highloss_rosetta["wt"] == res_wt)
                        & (df_highloss_rosetta["mt"] == res_mt)
                    ]
                ) / len(df[(df["wt"] == res_wt) & (df["mt"] == res_mt)])
            if dataset_key != "Rosetta" and dataset_key != "Xstal_vs_AF2":
                mat_loss_ml_exp[i, j] = df["mae_ml_exp"][
                    (df["wt"] == res_wt) & (df["mt"] == res_mt)
                ].mean()
                mat_var_exp[i, j] = df["score_exp_fermi"][
                    (df["wt"] == res_wt) & (df["mt"] == res_mt)
                ].var()
                if len(df[(df["wt"] == res_wt) & (df["mt"] == res_mt)]) > 0:
                    mat_highloss_count_exp[i, j] = len(
                        df_highloss_exp[
                            (df_highloss_exp["wt"] == res_wt)
                            & (df_highloss_exp["mt"] == res_mt)
                        ]
                    ) / len(df[(df["wt"] == res_wt) & (df["mt"] == res_mt)])

    # Make loss, var and count plots
    make_heatmap(dataset_key, mat_loss_ml_rosetta, aa_list, "loss_ml_rosetta")
    make_heatmap(dataset_key, mat_var_rosetta, aa_list, "var_rosetta")
    make_heatmap(
        dataset_key, mat_highloss_count_rosetta, aa_list, "highloss_count_rosetta"
    )
    if dataset_key != "Rosetta" and dataset_key != "Xstal_vs_AF2":
        make_heatmap(dataset_key, mat_loss_ml_exp, aa_list, "loss_ml_exp")
        make_heatmap(dataset_key, mat_var_exp, aa_list, "var_exp")
        make_heatmap(dataset_key, mat_highloss_count_exp, aa_list, "highloss_count_exp")


def make_heatmap(dataset_key, mat, aa_list, description):
    fig, ax = plt.subplots()
    fig.set_size_inches(10.0, 8.0)
    df_heat = pd.DataFrame(mat, columns=aa_list, index=aa_list)
    heatmap_loss = plt.pcolor(df_heat)
    cbar = plt.colorbar(heatmap_loss)
    cbar.set_label("Mean absolute error [kcal/mol]")
    ax.set_yticks(np.arange(0.5, len(df_heat.index), 1))
    ax.set_yticklabels(aa_list)
    ax.set_xticks(np.arange(0.5, len(df_heat.columns), 1))
    ax.set_xticklabels(aa_list)
    ax.set_xlabel("Variant amino acid")
    ax.set_ylabel("Wild-type amino acid")
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/heatmap_{description}.pdf",
        bbox_inches="tight",
    )
    plt.close()


def homology_plot(df, dataset_key):
    """
    Plots correlation coefficients to experimental ddGs based on homologous structures.
    Lines for both Rosetta and RaSP.
    """

    # Make new columns
    df["pdb"] = df["pdbid"].str[:4]
    df["bin"] = df["pdbid"].str[8:9].astype("int64")

    pdb_list = df["pdb"].unique()
    bin_list = df["bin"].unique()

    # Plot
    fig, ax_list = plt.subplots(nrows=len(pdb_list), ncols=1, sharex=True)
    fig.subplots_adjust(wspace=0.2)
    fig.tight_layout(pad=2, w_pad=5, h_pad=1)
    fig.text(0.5, -0.001, "Sequence identity", ha="center")
    fig.text(-0.0005, 0.5, "Pearson $\\rho$", va="center", rotation="vertical")

    # Define protein names
    name_dict = {
        "1LZ1": "Lysozyme",
        "2CI2": "Chymotrypsin inhibitor",
        "2RN2": "RNAse",
        "1BVC": "Myoglobin",
    }

    # Define vertical line and seq id positions
    vline_list = [100, 85, 70, 55, 40, 25]

    x_dict = {
        "1LZ1": [100, 98, 78, 62, 47, 33],
        "2CI2": [100, 91, 82, 40, 34],
        "2RN2": [100, 90, 72, 67, 43, 30],
        "1BVC": [100, 92, 81, 65, 43, 31],
    }

    # Lopp over proteins
    for i, pdb in enumerate(pdb_list):
        df_pdb = df[df["pdb"] == pdb]
        df_pdb = df_pdb.sort_values("bin")
        bin_list = df_pdb["bin"].unique()
        ax = ax_list[i]

        # Define x positions
        x = x_dict[pdb]

        # Loop over data points
        corr_ml = []
        corr_rosetta = []
        for i, binid in enumerate(bin_list):
            df_bin = df_pdb[df_pdb["bin"] == binid]
            df_bin_mean = df_bin.groupby("variant").mean()
            if len(df_bin_mean) != 0:
                corr_ml.append(
                    stats.pearsonr(df_bin_mean["score_ml"], df_bin_mean["score_exp"])[0]
                )
                corr_rosetta.append(
                    stats.pearsonr(
                        df_bin_mean["score_rosetta"], df_bin_mean["score_exp"]
                    )[0]
                )

        # Make scatter plot
        l1 = ax.plot(x, corr_rosetta, label="Rosetta", color="blue", marker=".")
        l2 = ax.plot(x, corr_ml, label="RaSP", color="orange", marker=".")
        anchored_text = AnchoredText(
            name_dict[pdb], loc="upper right", prop={"size": 8}
        )
        ax.add_artist(anchored_text)
        ax.set_xticks([92.5, 77.5, 62.5, 47.5, 32.5])
        ax.set_yticks([1.00, 0.75, 0.5, 0.25])
        ax.set_yticklabels(["1.00", "0.75", "0.50", "0.25"], fontsize="8")
        lns = l1 + l2
        labels = [l.get_label() for l in lns]

        # Define dashed vertical lines
        for vline in vline_list:
            ax.axvline(vline, color="lightgrey", linestyle="--")

    # Define x axis text and save fig
    ax.invert_xaxis()
    ax.set_xticklabels(["bin1", "bin2", "bin3", "bin4", "bin5"], fontsize="9")
    ax.text(vline_list[0] + 2.5, 0.05, "100%", color="dimgrey", fontsize="12")
    ax.text(vline_list[1] + 2.5, 0.05, "85%", color="dimgrey", fontsize="12")
    ax.text(vline_list[2] + 2.5, 0.05, "70%", color="dimgrey", fontsize="12")
    ax.text(vline_list[3] + 2.5, 0.05, "55%", color="dimgrey", fontsize="12")
    ax.text(vline_list[4] + 2.5, 0.05, "40%", color="dimgrey", fontsize="12")
    ax.text(vline_list[5] + 2.5, 0.05, "25%", color="dimgrey", fontsize="12")
    fig.legend(lns, labels, loc="upper center", ncol=2, prop={"size": 12})
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/homology_plot.pdf",
        bbox_inches="tight",
    )


def gnomAD_category(df):
    if df > 10**-2:
        return "gnomAD AF > 1e-2"

    elif df < 10**-2 and df > 10**-4:
        return "1e-2 > gnomAD AF > 1e-4"

    elif df < 10**-4:
        return "gnomAD AF < 1e-4"
    else:
        return np.nan


def category_back(df):
    if df != np.nan:
        return "Back"
    else:
        return np.nan


def intvl_95(np_array):
    d = []
    N = len(np_array)
    p = math.floor(0.95 * N)

    for i in range(N - p):
        d_i = np_array[i + p] - np_array[i]
        d.append(d_i)

    i = np.argmin(d)
    j = i + p

    r = (np_array[i], np_array[j])
    return np.around(r, decimals=2)


def plot_gnomad_clinvar():
    # Load data
    df = pd.read_csv(
        f"{os.path.dirname(os.getcwd())}/data/test/GnomAD_ClinVar/df_rasp_gnomad_clinvar.csv"
    )
    df["gnomAD_category"] = df.apply(lambda x: gnomAD_category(x["gnomad;AF"]), axis=1)
    df["category_back"] = df.apply(lambda x: category_back(x["RaSP;score_ml"]), axis=1)

    # Initialize
    y_RaSP = "RaSP;score_ml"
    x_back = "category_back"
    x_clinvar = "clinvar;Clinvar_signifiance"
    x_gnomad = "gnomAD_category"
    xlim = [-1, 7]

    # Data arrays
    benign_arr = np.sort(
        np.array(
            df["RaSP;score_ml"].values[
                np.where(df["clinvar;Clinvar_signifiance"].values == "benign")
            ]
        )
    )
    pathogenic_arr = np.sort(
        np.array(
            df["RaSP;score_ml"].values[
                np.where(df["clinvar;Clinvar_signifiance"].values == "pathogenic")
            ]
        )
    )

    gnomAD0_arr = np.sort(
        np.array(
            df["RaSP;score_ml"].values[
                np.where(df["gnomAD_category"].values == "gnomAD AF > 1e-2")
            ]
        )
    )
    gnomAD1_arr = np.sort(
        np.array(
            df["RaSP;score_ml"].values[
                np.where(df["gnomAD_category"].values == "1e-2 > gnomAD AF > 1e-4")
            ]
        )
    )
    gnomAD2_arr = np.sort(
        np.array(
            df["RaSP;score_ml"].values[
                np.where(df["gnomAD_category"].values == "gnomAD AF < 1e-4")
            ]
        )
    )

    # Plotting Raincloud plots of different distributions
    # Setting plot style
    sns.set(style="whitegrid", font_scale=1.5)

    # Setting up subplot figure
    fig, axs = plt.subplots(nrows=6, ncols=1, figsize=(10, 20))

    # Background RaSP distribution plotted as a raincloud
    pt.RainCloud(
        x=x_back,
        y=y_RaSP,
        data=df,
        order=["Back"],
        orient="h",
        palette=sns.color_palette("colorblind")[7:],
        bw=0.1,
        alpha=0.65,
        ax=axs[0],
    )

    # ClinVar 'benign' classified RaSP distribution plotted as a raincloud
    pt.RainCloud(
        x=x_clinvar,
        y=y_RaSP,
        data=df,
        order=["benign"],
        orient="h",
        palette=sns.color_palette("colorblind")[0:],
        bw=0.1,
        alpha=0.65,
        ax=axs[1],
    )

    cat_benign_median = np.nanmedian(
        df["RaSP;score_ml"].values[
            np.where(df["clinvar;Clinvar_signifiance"].values == "benign")
        ]
    )
    i_benign, j_benign = intvl_95(benign_arr)

    axs[1].annotate(
        "Median: "
        + str(np.around(cat_benign_median, decimals=2))
        + " "
        + "["
        + str(i_benign)
        + ";"
        + str(j_benign)
        + "]",
        (-0.5, 0.25),
        fontsize=11,
        color=sns.color_palette("colorblind")[0],
    )

    # ClinVar 'pathogenic' classified RaSP distribution plotted as a raincloud
    pt.RainCloud(
        x=x_clinvar,
        y=y_RaSP,
        data=df,
        order=["pathogenic"],
        orient="h",
        palette=sns.color_palette("colorblind")[1:],
        bw=0.1,
        alpha=0.65,
        ax=axs[2],
    )

    cat_patho_median = np.nanmedian(
        df["RaSP;score_ml"].values[
            np.where(df["clinvar;Clinvar_signifiance"].values == "pathogenic")
        ]
    )
    i_patho, j_patho = intvl_95(pathogenic_arr)

    axs[2].annotate(
        "Median: "
        + str(np.around(cat_patho_median, decimals=2))
        + " "
        + "["
        + str(i_patho)
        + ";"
        + str(j_patho)
        + "]",
        (-0.5, 0.25),
        fontsize=11,
        color=sns.color_palette("colorblind")[1],
    )

    # RaSP distribution for gnomAD AF_tot > 1e-2 plotted as a raincloud
    pt.RainCloud(
        x=x_gnomad,
        y=y_RaSP,
        data=df,
        order=["gnomAD AF > 1e-2"],
        palette=sns.color_palette("colorblind")[2:],
        orient="h",
        bw=0.1,
        alpha=0.65,
        ax=axs[3],
    )

    cat_0_median = np.nanmedian(
        df["RaSP;score_ml"].values[
            np.where(df["gnomAD_category"].values == "gnomAD AF > 1e-2")
        ]
    )
    i_cat0, j_cat0 = intvl_95(gnomAD0_arr)

    axs[3].annotate(
        "Median: "
        + str(np.around(cat_0_median, decimals=2))
        + " "
        + "["
        + str(i_cat0)
        + ";"
        + str(j_cat0)
        + "]",
        (-0.5, 0.25),
        fontsize=11,
        color=sns.color_palette("colorblind")[2],
    )

    # RaSP distribution for 1e-2 > gnomAD AF_tot > 1e-4 plotted as a raincloud
    pt.RainCloud(
        x=x_gnomad,
        y=y_RaSP,
        data=df,
        order=["1e-2 > gnomAD AF > 1e-4"],
        palette=sns.color_palette("colorblind")[3:],
        orient="h",
        bw=0.1,
        alpha=0.65,
        ax=axs[4],
    )

    cat_1_median = np.nanmedian(
        df["RaSP;score_ml"].values[
            np.where(df["gnomAD_category"].values == "1e-2 > gnomAD AF > 1e-4")
        ]
    )
    i_cat1, j_cat1 = intvl_95(gnomAD1_arr)

    axs[4].annotate(
        "Median: "
        + str(np.around(cat_1_median, decimals=2))
        + " "
        + "["
        + str(i_cat1)
        + ";"
        + str(j_cat1)
        + "]",
        (-0.5, 0.25),
        fontsize=11,
        color=sns.color_palette("colorblind")[3],
    )

    # RaSP distribution for gnomAD AF_tot < 1e-4 plotted as a raincloud
    pt.RainCloud(
        x=x_gnomad,
        y=y_RaSP,
        data=df,
        order=["gnomAD AF < 1e-4"],
        palette=sns.color_palette("colorblind")[4:],
        orient="h",
        bw=0.1,
        alpha=0.65,
        ax=axs[5],
    )

    cat_2_median = np.nanmedian(
        df["RaSP;score_ml"].values[
            np.where(df["gnomAD_category"].values == "gnomAD AF < 1e-4")
        ]
    )
    i_cat2, j_cat2 = intvl_95(gnomAD2_arr)

    axs[5].annotate(
        "Median: "
        + str(np.around(cat_2_median, decimals=2))
        + " "
        + "["
        + str(i_cat2)
        + ";"
        + str(j_cat2)
        + "]",
        (-0.5, 0.25),
        fontsize=11,
        color=sns.color_palette("colorblind")[4],
    )

    # Global setting for all subplots
    for ax in axs:
        ax.set_xlim(xlim)
        ax.set_xlabel("RaSP \u0394\u0394G [kcal/mol]", fontsize=12, labelpad=15)
        ax.set(ylabel=None)
        ax.tick_params(labelsize=12)

    fig.align_labels()
    fig.savefig(
        f"{os.path.dirname(os.getcwd())}/output/GnomAD_ClinVar/rasp_gnomad_clinvar.pdf",
        bbox_inches="tight",
    )



"""Implementation of the Practical Molecular Optimization benchmark

The Practical Molecular Optimization (PMO) benchmark [1] is a collection of
objective functions for molecular optimization. It is a superset of GuacaMol [2],
including additional tasks such as DRD2, JNK3, and GSK3Beta.

Our implementation heavily relies on the Therapeutics Data Commons [3], which
provides these black boxes.

Check the details in the respective papers.

References
----------
[1] Brown, Nathan, Marco Fiscato, Marwin H.S. Segler, and Alain C. Vaucher.
    “GuacaMol: Benchmarking Models for de Novo Molecular Design.”
    Journal of Chemical Information and Modeling 59, no. 3 (March 25, 2019):
    1096-1108. https://doi.org/10.1021/acs.jcim.8b00839.

[2] Gao, Wenhao, Tianfan Fu, Jimeng Sun, and Connor W. Coley.
    “Sample Efficiency Matters: A Benchmark for Practical Molecular Optimization,” 2022.
    https://openreview.net/forum?id=yCZRdI0Y7G.

[3] Huang, Kexin, Tianfan Fu, Wenhao Gao, Yue Zhao, Yusuf Roohani, Jure Leskovec,
    Connor W Coley, Cao Xiao, Jimeng Sun, and Marinka Zitnik.
    “Therapeutics Data Commons: Machine Learning Datasets and Tasks for Drug Discovery and Development.”
    Proceedings of Neural Information Processing Systems, NeurIPS Datasets and Benchmarks, 2021.

"""

from typing import Literal, Union

from poli.objective_repository import (
    DRD2ProblemFactory,
    GSK3BetaProblemFactory,
    JNK3ProblemFactory,
)

from .guacamol import GuacaMolGoalDirectedBenchmark


class PMOBenchmark(GuacaMolGoalDirectedBenchmark):
    """Practical Molecular Optimization benchmark

    The Practical Molecular Optimization (PMO) benchmark [1] is a
    collection of objective functions for molecular optimization.
    It is a superset of GuacaMol [2], including additional tasks such
    as DRD2, JNK3, and GSK3Beta.

    Our implementation heavily relies on the Therapeutics Data Commons [3],
    which provides these black boxes.

    Check the details in the respective papers.

    References
    ----------
    [1] Brown, Nathan, Marco Fiscato, Marwin H.S. Segler, and Alain C. Vaucher.
        “GuacaMol: Benchmarking Models for de Novo Molecular Design.”
        Journal of Chemical Information and Modeling 59, no. 3 (March 25, 2019):
        1096-1108. https://doi.org/10.1021/acs.jcim.8b00839.

    [2] Gao, Wenhao, Tianfan Fu, Jimeng Sun, and Connor W. Coley.
        “Sample Efficiency Matters: A Benchmark for Practical Molecular Optimization,” 2022.
        https://openreview.net/forum?id=yCZRdI0Y7G.

    [3] Huang, Kexin, Tianfan Fu, Wenhao Gao, Yue Zhao, Yusuf Roohani, Jure Leskovec,
        Connor W Coley, Cao Xiao, Jimeng Sun, and Marinka Zitnik.
        “Therapeutics Data Commons: Machine Learning Datasets and Tasks for Drug Discovery and Development.”
        Proceedings of Neural Information Processing Systems, NeurIPS Datasets and Benchmarks, 2021.
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"],
        seed: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        parallelize: bool = False,
        num_workers: Union[int, None] = None,
        evaluation_budget: int = float("inf"),
    ) -> None:
        super().__init__(
            string_representation=string_representation,
            seed=seed,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

        self.problem_factories.extend(
            [
                JNK3ProblemFactory(),
                GSK3BetaProblemFactory(),
                DRD2ProblemFactory(),
            ]
        )



"""
Implements the JNK3 task using the TDC oracles [1].

When using this task, we recommend you cite [1, 2, 3, 4].

References
----------
[1] Artificial intelligence foundation for therapeutic science.
    Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022).
    https://doi.org/10.1038/s41589-022-01131-2
[2] Multi-objective de novo drug design with conditional graph generative models.
    Li, Y., Zhang, L., Liu, Z.
    Journal of cheminformatics 10.1 (2018).
[3] Multi-objective molecule generation using interpretable substructures.
    Jin, Wengong, Regina Barzilay, and Tommi Jaakkola.
    ICML. 2020.
    https://people.csail.mit.edu/tommi/papers/JBJ_ICML2020b.pdf
[4] ExCAPE-DB: an integrated large scale dataset facilitating Big Data analysis in chemogenomics.
    Jiangming, Sun, et al.
    Journal of cheminformatics 9.1 (2017).
"""

from __future__ import annotations

from typing import Literal

import numpy as np
import selfies as sf

from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.chemistry.tdc_black_box import TDCBlackBox
from poli.core.chemistry.tdc_problem import TDCProblem
from poli.core.problem import Problem
from poli.core.util.chemistry.string_to_molecule import translate_smiles_to_selfies
from poli.core.util.seeding import seed_numpy, seed_python


class JNK3BlackBox(TDCBlackBox):
    """
    A black box giving access to the JNK3 task, provided by TDC [1].

    When using this task, we recommend you cite [1, 2, 3, 4].

    Parameters
    ----------
    string_representation : Literal["SMILES", "SELFIES"], optional
        A string (either "SMILES" or "SELFIES") specifying which
        molecule representation you plan to use.
    alphabet : list[str] | None, optional
        The alphabet to be used for the SMILES or SELFIES representation.
        It is common that the alphabet depends on the dataset used, so
        it is recommended to pass it as an argument. Default is None.
    max_sequence_length : int, optional
        The maximum length of the sequence. Default is infinity.
    batch_size : int, optional
        The batch size for simultaneous execution, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize execution, by default False.
    num_workers : int, optional
        The number of workers for parallel execution, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.
    force_isolation: bool, optional
        Whether to force the isolation of the black box. Default is False.

    Attributes
    ----------
    oracle_name : str
        The name of the oracle.

    Methods
    -------
    __init__(self, info, batch_size=None, parallelize=False, num_workers=None, from_smiles=True)
        Initializes a new instance of the DRD3BlackBox class.

    References
    ----------
    [1] Artificial intelligence foundation for therapeutic science.
        Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
    [2] Multi-objective de novo drug design with conditional graph generative models.
        Li, Y., Zhang, L., Liu, Z.
        Journal of cheminformatics 10.1 (2018).
    [3] Multi-objective molecule generation using interpretable substructures.
        Jin, Wengong, Regina Barzilay, and Tommi Jaakkola.
        ICML. 2020.
        https://people.csail.mit.edu/tommi/papers/JBJ_ICML2020b.pdf
    [4] ExCAPE-DB: an integrated large scale dataset facilitating Big Data analysis in chemogenomics.
        Jiangming, Sun, et al.
        Journal of cheminformatics 9.1 (2017).
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        force_isolation: bool = False,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):
        super().__init__(
            oracle_name="JNK3",
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            force_isolation=force_isolation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="jnk3",
            max_sequence_length=self.max_sequence_length,
            aligned=False,
            fixed_length=False,
            deterministic=True,  # ?
            alphabet=self.alphabet,  # TODO: add alphabet once we settle for one for SMLIES/SELFIES.
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class JNK3ProblemFactory(AbstractProblemFactory):
    """
    Factory class for creating JNK3 docking problems.

    When using this task, we recommend you cite [1, 2, 3, 4].

    Methods
    ------
    create:
        Creates a JNK3 problem.

    References
    ----------
    [1] Artificial intelligence foundation for therapeutic science.
        Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
    [2] Multi-objective de novo drug design with conditional graph generative models.
        Li, Y., Zhang, L., Liu, Z.
        Journal of cheminformatics 10.1 (2018).
    [3] Multi-objective molecule generation using interpretable substructures.
        Jin, Wengong, Regina Barzilay, and Tommi Jaakkola.
        ICML. 2020.
        https://people.csail.mit.edu/tommi/papers/JBJ_ICML2020b.pdf
    [4] ExCAPE-DB: an integrated large scale dataset facilitating Big Data analysis in chemogenomics.
        Jiangming, Sun, et al.
        Journal of cheminformatics 9.1 (2017).
    """

    def create(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates a JNK3 problem.

        Parameters
        ----------
        string_representation : str, optional
            The string representation of the molecules. Must be either 'SMILES' or 'SELFIES'. Default is 'SMILES'.
        alphabet : list[str] | None, optional
            The alphabet to be used for the SMILES or SELFIES representation.
            It is common that the alphabet depends on the dataset used, so
            it is recommended to pass it as an argument. Default is None.
        max_sequence_length : int, optional
            The maximum length of the sequence. Default is infinity.
        seed : int, optional
            Seed for random number generators. If None, no seed is set.
        batch_size : int, optional
            Number of molecules to process in parallel. If None, the default batch size is used.
        parallelize : bool, optional
            Whether to parallelize the docking process. Default is False.
        num_workers : int, optional
            Number of worker processes to use for parallelization. If None, the number of available CPU cores is used.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        force_isolation: bool, optional
            Whether to force the isolation of the black box. Default is False.

        Returns
        -------
        problem : Problem
            A problem instance containing the black box, and an initial value x0.

        Raises
        ------
        ValueError
            If the string_representation is not 'SMILES' or 'SELFIES'.
        """
        # We start by seeding the RNGs
        if seed is not None:
            seed_numpy(seed)
            seed_python(seed)

        # We check whether the string representation is valid
        if string_representation.upper() not in ["SMILES", "SELFIES"]:
            raise ValueError(
                "Missing required keyword argument: string_representation: str. "
                "String representation must be either 'SMILES' or 'SELFIES'."
            )

        f = JNK3BlackBox(
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            force_isolation=force_isolation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

        # Initial example (from the TDC docs)
        x0_smiles = "CC(C)(C)[C@H]1CCc2c(sc(NC(=O)COc3ccc(Cl)cc3)c2C(N)=O)C1"
        x0_selfies = translate_smiles_to_selfies([x0_smiles])[0]

        if string_representation.upper() == "SMILES":
            x0 = np.array([list(x0_smiles)])
        else:
            x0 = np.array([list(sf.split_selfies(x0_selfies))])

        jnk3_problem = TDCProblem(
            black_box=f,
            x0=x0,
        )

        return jnk3_problem



from pathlib import Path

import numpy as np
import pytest

from poli import objective_factory

THIS_DIR = Path(__file__).parent.resolve()

HOME_DIR = Path().home().resolve()
PATH_TO_FOLDX_FILES = HOME_DIR / "foldx"
if not PATH_TO_FOLDX_FILES.exists():
    pytest.skip("FoldX is not installed. ", allow_module_level=True)

if not (PATH_TO_FOLDX_FILES / "foldx").exists():
    pytest.skip("FoldX is not compiled. ", allow_module_level=True)


@pytest.mark.poli__protein
def test_running_foldx_stability():
    """
    Testing whether we can register the logp problem
    if biopython and python-levenshtein are installed.
    """
    problem = objective_factory.create(
        name="foldx_stability",
        wildtype_pdb_path=THIS_DIR / "101m_Repair.pdb",
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)

    assert np.isclose(y0, 32.4896).all()


@pytest.mark.poli__protein
def test_running_foldx_sasa():
    """
    Testing whether we can register the logp problem
    if biopython and python-levenshtein are installed.
    """
    problem = objective_factory.create(
        name="foldx_sasa",
        wildtype_pdb_path=THIS_DIR / "101m_Repair.pdb",
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)

    assert np.isclose(y0, 8411.45578009).all()


@pytest.mark.poli__protein
def test_running_foldx_stability_and_sasa():
    """
    Testing whether we can register the logp problem
    if biopython and python-levenshtein are installed.
    """
    problem = objective_factory.create(
        name="foldx_stability_and_sasa",
        wildtype_pdb_path=THIS_DIR / "101m_Repair.pdb",
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)

    assert np.isclose(y0[:, 0], 32.4896).all()
    assert np.isclose(y0[:, 1], 8411.45578009).all()


@pytest.mark.poli__protein
def test_registering_foldx_stability_and_sasa_with_verbose_output():
    """
    Testing whether the foldx output is printed.
    """
    problem = objective_factory.create(
        name="foldx_stability_and_sasa",
        wildtype_pdb_path=THIS_DIR / "101m_Repair.pdb",
        verbose=True,
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)

    assert np.isclose(y0[:, 0], 32.4896).all()
    assert np.isclose(y0[:, 1], 8411.45578009).all()


@pytest.mark.poli__protein
@pytest.mark.slow()
def test_foldx_from_non_repaired_file():
    """
    In this test, we check whether foldx properly
    repairs a file if it doesn't contain _Repair.

    TODO: mock the behavior of the repair function
    inside the foldx interface. Otherwise, this test
    takes 4min to run.
    """
    wildtype_pdb_path = THIS_DIR / "3ned.pdb"
    problem = objective_factory.create(
        name="foldx_stability",
        wildtype_pdb_path=wildtype_pdb_path,
        eager_repair=True,
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)

    assert np.isclose(y0, 32.6135).all()


@pytest.mark.poli__protein
def test_foldx_from_repaired_file():
    """
    In this test, we check whether no repair is
    performed if the file already contains _Repair.
    """
    wildtype_pdb_path = THIS_DIR / "101m_Repair.pdb"
    problem = objective_factory.create(
        name="foldx_stability",
        wildtype_pdb_path=wildtype_pdb_path,
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)

    assert np.isclose(y0, 32.4896).all()



"""Utilities for downloading files from GitHub repositories.

This module requires the PyGithub package, install it with:

        pip install PyGithub

Most of this code was taken and adapted from:
https://gist.github.com/pdashford/2e4bcd4fc2343e2fd03efe4da17f577d?permalink_comment_id=4274705#gistcomment-4274705
"""

import base64
import os
from pathlib import Path

from github import BadCredentialsException, Github, GithubException
from github.ContentFile import ContentFile
from github.Repository import Repository


def get_sha_for_tag(repository: Repository, tag: str) -> str:
    """
    Returns a commit PyGithub object for the specified repository and tag.

    Parameters
    ----------
    repository : Repository
        The repository.
    tag : str
        The tag.

    Returns
    -------
    commit_sha: str
        The commit SHA for the specified repository and tag.

    Raises
    ------
    ValueError
        If no tag or branch exists with the specified name.

    Examples
    --------
    >>> from github import Github
    >>> from github.Repository import Repository
    >>> github = Github()
    >>> repository = github.get_repo("rdkit/rdkit")
    >>> get_sha_for_tag(repository, "Release_2023_09")
    '068441957858f786c227825d90eb2c43f4f2b000'
    """
    branches = repository.get_branches()
    matched_branches = [match for match in branches if match.name == tag]
    if matched_branches:
        return matched_branches[0].commit.sha

    tags = repository.get_tags()
    matched_tags = [match for match in tags if match.name == tag]
    if not matched_tags:
        raise ValueError("No Tag or Branch exists with that name")

    return matched_tags[0].commit.sha


def download_file_from_github_repository(
    repository_name: str,
    file_path_in_repository: str,
    download_path_for_file: str,
    tag: str = "master",
    commit_sha: str = None,
    exist_ok: bool = False,
    parent_folders_exist_ok: bool = True,
    verbose: bool = False,
    strict: bool = True,
) -> None:
    """
    Download a file from a Github repository.

    Parameters
    ----------
    repository_name: str, required
        The name of the repository (i.e. "user/repo")
    file_path_in_repository: str, required
        path to file in repo
    download_path: str, required
        path to download to
    tag: str, optional
        tag or branch to download, defaults to master
    sha: str, optional
        sha of commit to download, overwrites tag if specified
    exists_ok: bool, optional
        whether to overwrite existing files
    parent_folders_exist_ok: bool, optional
        whether to create parent folders if they do not exist
    verbose: bool, optional
        whether to print progress
    strict: bool, optional
        whether to raise exceptions on errors

    Warnings
    --------

    This function will use an environment variable called
    GITHUB_TOKEN_FOR_POLI if it exists. If it does not exist,
    it will try to download without it. Note that the rate limit
    is 60 requests per hour for anonymous requests.

    To create a GitHub token like this, follow the instructions here:
    https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens#creating-a-fine-grained-personal-access-token
    """
    github = Github(login_or_token=os.environ.get("GITHUB_TOKEN_FOR_POLI"))

    try:
        repository = github.get_repo(repository_name)
    except BadCredentialsException as e:
        raise ValueError(
            "Your token has likely expired. Please set a new token following the instructions here: https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/managing-your-personal-access-tokens#creating-a-fine-grained-personal-access-token"
        ) from e

    if commit_sha is None:
        commit_sha = get_sha_for_tag(repository, tag)

    _download_file_from_github_repo(
        repository=repository,
        commit_sha=commit_sha,
        file_path_in_repository=file_path_in_repository,
        download_path_for_file=download_path_for_file,
        exist_ok=exist_ok,
        parent_folders_exist_ok=parent_folders_exist_ok,
        verbose=verbose,
        strict=strict,
    )


def _download_file_from_github_repo(
    repository: Repository,
    commit_sha: str,
    file_path_in_repository: str,
    download_path_for_file: str,
    exist_ok: bool = False,
    parent_folders_exist_ok: bool = True,
    verbose: bool = False,
    strict: bool = True,
) -> None:
    """
    Download a file from a GitHub repository.

    Parameters
    ----------
    repository : Repository
        The GitHub repository object.
    commit_sha : str
        The commit SHA of the file in the repository.
    file_path_in_repository : str
        The path of the file in the repository.
    download_path_for_file : str
        The path where the file will be downloaded.
    exist_ok : bool, optional
        If False and the download path already exists, a
        FileExistsError will be raised.
        If True, the download will proceed even if the path
        already exists. (default is False)
    parent_folders : bool, optional
        If True, create the parent folders for the download path
        if they do not exist. (default is True)
    verbose : bool, optional
        If True, print the progress of the download. (default is False)
    strict : bool, optional
        If True, raise an exception if there is an error during the download.
        If False, print an error message and continue. (default is True)
    """
    if os.path.exists(download_path_for_file):
        if not exist_ok:
            raise FileExistsError("Path already exists: %s")

    if not isinstance(download_path_for_file, Path):
        download_path_for_file = Path(download_path_for_file)

    download_path_for_file.parent.mkdir(parents=True, exist_ok=parent_folders_exist_ok)

    file_content = repository.get_contents(file_path_in_repository, ref=commit_sha)

    if isinstance(file_content, ContentFile):
        _save_file_content_from_github(
            file_content, download_path_for_file, strict=strict, verbose=verbose
        )
    elif isinstance(file_content, list):
        for file_ in file_content:
            _download_file_from_github_repo(
                repository=repository,
                commit_sha=commit_sha,
                file_path_in_repository=file_.path,
                download_path_for_file=download_path_for_file / file_.name,
                exist_ok=exist_ok,
                parent_folders_exist_ok=parent_folders_exist_ok,
                verbose=verbose,
                strict=strict,
            )
    else:
        raise ValueError("Expected ContentFile or list of ContentFile")


def _save_file_content_from_github(
    file_content: ContentFile,
    download_path_for_file: Path,
    strict: bool = True,
    verbose: bool = False,
) -> None:
    if verbose:
        print(f"poli 🧪: Downloading {file_content.path}")
    try:
        if not isinstance(file_content, ContentFile):
            raise ValueError("Expected ContentFile")

        with open(download_path_for_file, "wb") as file_out:
            if file_content.content:
                file_data = base64.b64decode(file_content.content)
                file_out.write(file_data)
            elif file_content.download_url:
                import requests

                file_out.write(requests.get(file_content.download_url).content)

    except (GithubException, IOError, ValueError) as exc:
        if strict:
            raise exc
        else:
            print("Error processing %s: %s", file_content.path, exc)



"""Default observer that just does nothing."""

from poli.core.util.abstract_observer import AbstractObserver


class DefaultObserver(AbstractObserver):
    """
    This observer serves as a place-holder if the user does not want to use observers.
    """

    def observe(self, *args, **kwargs) -> None:
        pass

    def initialize_observer(self, *args, **kwargs) -> object:
        return None



from __future__ import annotations

import configparser
import importlib
import logging
import shutil
import subprocess
from pathlib import Path

from poli.core.registry import _ISOLATED_FUNCTION_SCRIPT_LOCATION, _OBSERVER
from poli.core.util.inter_process_communication.process_wrapper import ProcessWrapper

from .external_function import ExternalFunction

HOME_DIR = Path.home().resolve()
(HOME_DIR / ".poli_objectives").mkdir(exist_ok=True)

config_file = str(HOME_DIR / ".poli_objectives" / "config.rc")
config = configparser.ConfigParser(defaults={_OBSERVER: ""})
ls = config.read(config_file)


def load_config():
    """Loads the configuration file containing which objectives are registered.

    Returns
    -------
    config : configparser.ConfigParser
        The configuration file.

    """
    HOME_DIR = Path.home().resolve()
    config_file = str(HOME_DIR / ".poli_objectives" / "config.rc")
    config = configparser.ConfigParser(defaults={_OBSERVER: ""})
    _ = config.read(config_file)

    return config


def __read_env_name(environment_file: Path) -> str:
    with open(environment_file, "r") as f:
        # This is a really crude way of doing this,
        # but it works. We should probably use a
        # yaml parser instead, but the idea is to keep
        # the dependencies to a minimum.
        yml = f.read()
        lines = yml.split("\n")
        conda_env_name_line = lines[0]
        assert conda_env_name_line.startswith("name:"), (
            "The first line of the environment.yml file "
            "should be the name of the environment"
        )
        env_name = lines[0].split(":")[1].strip()
    return env_name


def __create_conda_env(environment_file: Path, quiet: bool = False):
    env_name = __read_env_name(environment_file=environment_file)

    # 1. create the environment from the yaml file
    if not quiet:
        print(f"poli 🧪: creating environment {env_name} from {environment_file}")
    try:
        subprocess.run(
            " ".join(["conda", "env", "create", "-f", str(environment_file)]),
            shell=True,
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        if "already exists" in e.stderr.decode():
            if not quiet:
                print(f"poli 🧪: {env_name} already exists.")
        else:
            raise e


def __register_isolated_function_from_repository(
    name: str, quiet: bool = False
) -> None:
    """Registers a problem from the repository.

    This function takes a problem name, and registers it. The problem name
    corresponds to a folder inside the objective_repository folder. The
    function will:
    1. create the environment from the yaml file
    2. run the file from said enviroment (since we can't
    import the factory: it may have dependencies that are
    not installed)

    Parameters
    ----------
    name : str
        The name of the problem to be registered.
    quiet : bool, optional
        If True, we squelch the feedback about environment creation and
        problem registration, by default False.
    """
    # Load up the environment name
    PATH_TO_REPOSITORY = (
        Path(__file__).parent.parent.parent.parent / "objective_repository"
    ).resolve()

    name_without_isolated = name.replace("__isolated", "")
    environment_file = PATH_TO_REPOSITORY / name_without_isolated / "environment.yml"
    isolated_file = PATH_TO_REPOSITORY / name_without_isolated / "isolated_function.py"

    __register_isolated_file(
        environment_file=environment_file,
        isolated_file=isolated_file,
        name_for_show=name,
        quiet=quiet,
    )


def __register_isolated_function_from_core(name: str, quiet: bool = False) -> None:
    """
    Registers an isolated function from the core package.

    At the moment, we only have one isolated function
    available in the core package, which is the TDCIsolatedFunction.

    Parameters
    ----------
    name : str
        The name of the isolated function to register. At the moment,
        the only available isolated function is "tdc__isolated".
    quiet : bool, optional
        If True, we squelch the feedback about environment creation and
        problem registration, by default False.
    """
    ROOT_DIR_OF_POLI_PACKAGE = Path(__file__).parent.parent.parent.parent
    if name == "tdc__isolated":
        environment_file = (
            ROOT_DIR_OF_POLI_PACKAGE / "core" / "chemistry" / "environment.yml"
        )
        isolated_file = (
            ROOT_DIR_OF_POLI_PACKAGE / "core" / "chemistry" / "tdc_isolated_function.py"
        )
        __register_isolated_file(
            environment_file=environment_file,
            isolated_file=isolated_file,
            name_for_show="TDC isolated function",
            quiet=quiet,
        )
    else:
        raise NotImplementedError(
            "The only core isolated function available is the "
            "TDCIsolatedFunction (i.e. tdc__isolated)."
        )


def __run_file_in_env(env_name: str, file_path: Path):
    """
    Runs a file from a given conda env.
    """
    command = " ".join(["conda", "run", "-n", env_name, "python", str(file_path)])
    try:
        subprocess.run(command, check=True, shell=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Found error when running {file_path} from environment {env_name}: \n"
            f"{e.stderr.decode()}"
        ) from e


def __register_isolated_file(
    environment_file: Path,
    isolated_file: Path,
    name_for_show: str = None,
    quiet: bool = False,
):
    """
    Creates the conda environment in the environment file and
    runs the isolated file.
    """
    # 1. Create the conda env.
    __create_conda_env(environment_file, quiet=quiet)
    env_name = __read_env_name(environment_file)

    # 2. Running the file
    if not quiet:
        if name_for_show:
            print(
                f"poli 🧪: running registration of {name_for_show} from environment {env_name}"
            )
        else:
            print(f"poli 🧪: running {isolated_file} from environment {env_name}")

    __run_file_in_env(env_name, isolated_file)


def register_isolated_function(name: str, quiet: bool = False):
    """Registers the objective function if it is available in the repository.

    If the objective function is not available in the repository,
    then we raise an error. If it is available, then we ask the
    user for confirmation to register it. If the user confirms,
    then we register it. Otherwise, we raise an error.

    Parameters
    ----------
    name : str
        The name of the objective function. This corresponds to
        the folder name inside the objective repository. If the
        name contains a `__isolated`, then it is assumed
        that the name refers to an internal file called
        `isolated_function.py`. An exception to this
        is "tdc__isolated", which registers the
        TDCIsolatedFunction.
    quiet : bool, optional
        If True, we squelch the messages giving feedback about the creation process.
        By default, it is False.
    """
    config = load_config()
    if name not in config:
        # Register problem

        # Two cases:
        # (i) some of the isolated functions are not alongside
        # their black boxes and problem factories, but are rather inside
        # the core of poli. For now, the only case is tdc, but more may
        # come in the future.
        #
        # (ii) the isolated function is in the repository, living alongside
        # the black box and the problem factory.
        if name == "tdc__isolated":
            logging.debug(
                "poli 🧪: Registered the isolated function from the repository."
            )
            __register_isolated_function_from_core(name, quiet=quiet)
            config = load_config()
        else:
            logging.debug(
                "poli 🧪: Registered the isolated function from the repository."
            )
            __register_isolated_function_from_repository(name, quiet=quiet)
            # Refresh the config
            config = load_config()


def __create_function_as_isolated_process(
    name: str,
    quiet: bool = False,
    **kwargs_for_isolated_function,
) -> ExternalFunction:
    """Creates the objective function as an isolated process.

    If the problem is registered, we create it as an isolated
    process. Otherwise, we raise an error. That is, this function
    expects the problem to be registered.

    Parameters
    ----------
    name : str
        The name of the objective function.
    seed : int, optional
        The seed value for random number generation.
    quiet : bool, optional
        If True, we squelch the messages giving feedback about the creation process.
        By default, it is False.
    **kwargs_for_factory : dict, optional
        Additional keyword arguments for the factory.
    """
    config = load_config()
    if name not in config:
        raise ValueError(
            f"Objective function '{name.replace('__isolated', '')}' is not registered. "
        )

    if not quiet:
        print(
            f"poli 🧪: Starting the function {name.replace('__isolated', '')} as an isolated process."
        )

    seed = kwargs_for_isolated_function.get("seed", None)
    process_wrapper = ProcessWrapper(
        config[name][_ISOLATED_FUNCTION_SCRIPT_LOCATION], **kwargs_for_isolated_function
    )
    seed = kwargs_for_isolated_function.get("seed", None)
    # TODO: add signal listener that intercepts when proc ends
    # wait for connection from objective process
    # TODO: potential (unlikely) race condition! (process might try to connect before listener is ready!)
    # TODO: We could be sending all the kwargs for the black box here.
    process_wrapper.send(("SETUP", seed, kwargs_for_isolated_function))

    msg_type, *msg = process_wrapper.recv()
    if msg_type == "SETUP":
        # Then the instance of the black box
        # was correctly set-up.
        pass
    elif msg_type == "EXCEPTION":
        e, tb = msg
        print(tb)
        raise e
    else:
        raise ValueError(
            f"Internal error: received {msg_type} when expecting SETUP or EXCEPTION"
        )

    f = ExternalFunction(process_wrapper)

    return f


def instance_function_as_isolated_process(
    name: str,
    quiet: bool = False,
    **kwargs_for_black_box,
) -> ExternalFunction:
    # Check if the user has conda installed
    if shutil.which("conda") is None:
        raise RuntimeError(
            "Conda is not installed/not in the PATH. For poli's isolation mechanisms to work, \n"
            "we need conda to be installed.\n"
            "If you are not interested in using conda, you can install all the \n"
            "relevant dependencies for black boxes using pip and optional arguments.\n"
            "Check the documentation of the black box you are interested in for more information.\n"
            "https://machinelearninglifescience.github.io/poli-docs/."
        )

    # Register the problem if it hasn't been registered.
    register_isolated_function(name=name, quiet=quiet)

    f = __create_function_as_isolated_process(
        name=name,
        quiet=quiet,
        **kwargs_for_black_box,
    )

    return f


def get_inner_function(
    isolated_function_name: str,
    class_name: str,
    module_to_import: str,
    force_isolation: bool = False,
    quiet: bool = False,
    **kwargs,
):
    """Utility for creating an instance of an inner isolated function.

    This function is used in almost every single black box that requires isolation,
    and it abstracts away the logic of trying to import the relevant AbstractIsolatedLogic
    class from the sibling isolated_function.py file of each register.py.

    Parameters
    ----------
    isolated_function_name : str
        The name of the isolated function to be created (e.g. "foldx_stability__isolated").
    class_name : str
        The name of the class to be imported from the isolated function file (e.g. FoldXStabilityIsolatedLogic).
    module_to_import : str
        The full name of the module to import the class from (e.g.
        "poli.objective_repository.foldx_stability.isolated_function").
    force_isolation : bool, optional
        If True, then the function is forced to run in isolation, even if the module can be imported.
    quiet : bool, optional
        If True, we squelch the messages giving feedback about the creation process.
        By default, it is False.
    **kwargs : dict
        Additional keyword arguments for the isolated function.
    """
    if not force_isolation:
        try:
            module = importlib.import_module(module_to_import)
            InnerFunctionClass = getattr(module, class_name)
            inner_function = InnerFunctionClass(**kwargs)
        except ImportError:
            inner_function = instance_function_as_isolated_process(
                name=isolated_function_name, quiet=quiet, **kwargs
            )
    else:
        inner_function = instance_function_as_isolated_process(
            name=isolated_function_name, quiet=quiet, **kwargs
        )
    return inner_function



"""
Module that wraps utility functions for interprocess communication.
"""

import logging
import subprocess
import time
from multiprocessing.connection import Client, Listener
from pathlib import Path
from uuid import uuid4


def get_connection(port: int, password: str) -> Client:
    """
    Get a connection to a server.

    Parameters
    ----------
    port : int
        The port number to connect to.

    password : str
        The password for authentication.

    Returns
    -------
    Client
        The client object representing the connection to the server.

    Raises
    ------
    EOFError
        If the host process is not ready yet.
    ConnectionRefusedError
        If the connection is refused by the server.

    Notes
    -----
    This function attempts to establish a connection to a server using the given port and password.
    It retries the connection up to two times before raising an exception.
    """
    address = ("", port)
    retries = 2
    while retries > 0:
        time.sleep(1)  # wait a second and then try to make a connection
        try:
            # if we manage to establish a connection we exit the function
            return Client(address, authkey=password.encode())
        # maybe the host process isn't ready yet
        except EOFError:
            pass
        except ConnectionRefusedError:
            pass
        retries -= 1
    # when we get here, e must have been instantiated
    logging.fatal("Could not connect to host process.")
    raise ConnectionError("Could not connect to host process.")


class ProcessWrapper:
    def __init__(self, run_script, **kwargs_for_factory):
        """
        Initialize the connection for inter process communication.

        Parameters
        ----------
        run_script : str
            The run script to execute. The run script should accept a port and a password as arguments.

        **kwargs_for_factory : dict
            Additional keyword arguments to be passed to the run script.
            These will be passed to the inner objective factory.

        Notes
        -----
        This class sets up a server for inter process communication. It generates a password for authentication,
        creates a listener on a random port, and starts a subprocess to execute the run script. The run script
        is expected to take the port and password as arguments, as well as any other arguments passed by the user
        when calling `objective_factory.create`.

        The `kwargs_for_factory` dictionary is used to convert the keyword arguments into a string format that can
        be passed to the run script. The string format is determined based on the type of the argument.
        """
        address = ("", 0)  # ('localhost', 6000)
        self.password = _generate_password()
        self.listener = Listener(address, authkey=self.password.encode())

        # TODO: very hacky way to read out the socket! (but the listener is not very cooperative)
        self.port = self.listener._listener._socket.getsockname()[1]
        # here is a VERY crucial step
        # we expect the shell script to take port and password as arguments, as well as other arguments passed by the user
        # when calling objective_factory.create
        # TODO: This is a very silly way to handle communication between processes,
        # and it is also very dangerous, because the user can pass arbitrary arguments
        # to the shell script. We should instead use a proper IPC library.
        # TODO: if we decide to pass this information in the set-up phase (instead
        # of here), we can remove this.
        string_for_kwargs = ""
        for key, value in kwargs_for_factory.items():
            if isinstance(value, str):
                string_for_kwargs += f"--{key}={str(value)} "
            elif isinstance(value, Path):
                string_for_kwargs += f"--{key}={str(value)} "
            elif isinstance(value, bool):
                string_for_kwargs += f"--{key}=bool:{str(value)} "
            elif isinstance(value, int):
                string_for_kwargs += f"--{key}=int:{str(value)} "
            elif isinstance(value, float):
                string_for_kwargs += f"--{key}=float:{str(value)} "
            elif isinstance(value, list):
                string_for_kwargs += (
                    f"--{key}=list:{','.join([str(v) for v in value])} "
                )
            elif value is None:
                string_for_kwargs += f"--{key}=none:None "

        self.run_script = run_script
        self.proc = subprocess.Popen(
            [run_script, str(self.port), self.password, string_for_kwargs],
            stdout=None,
            stderr=None,
        )

        self.conn = self.listener.accept()  # wait for the process to connect

    def send(self, *args):
        """
        Send data through the connection.

        Parameters
        ----------
        *args : tuple
            The data to be sent.

        Returns
        -------
        object
            The return value of the send operation.

        """
        return self.conn.send(*args)

    def recv(self):
        """
        Receive data from the connection.

        Returns
        -------
        object
            The received data.
        """
        return self.conn.recv()

    def close(self):
        """
        Closes the connection.
        """
        # TODO: potentially dangerous to wait here!
        self.proc.wait()  # wait for objective function process to finish
        self.listener.close()

    def __str__(self) -> str:
        return f"ProcessWrapper(port={self.port}, script_location={self.run_script})"

    def __repr__(self) -> str:
        return f"<{self.__str__()}>"


def _generate_password() -> str:
    """
    Generates a random password using uuid4.

    Returns
    -------
    password: str
        The generated password.
    """
    return str(uuid4())



def pytest_collection_modifyitems(items, config):
    for item in items:
        markers = item.iter_markers()
        is_unmarked = True
        for marker in markers:
            if marker.name != "parametrize":
                is_unmarked = False
                break

        if is_unmarked:
            item.add_marker("unmarked")



"""Utility model to check if the input to an "aligned"
problem is indeed aligned.
"""

import numpy as np


def is_aligned_input(x: np.ndarray, maximum_sequence_length: int = None) -> bool:
    """Utility function to check if the input to an "aligned"
    problem is indeed aligned.

    Our definition of alignment goes as follows:
    - The input is of shape [b, L], where L is a defined maximum
      length.
    - If the input is of shape [b,], then the lengths of all
      elements inside the batch are the same.
    """
    if maximum_sequence_length == np.inf:
        maximum_sequence_length = None

    # If the array is 2-dimensional, we only need to check
    # if the second dimension is equal to the maximum sequence length
    # (when provided)
    if x.ndim == 2:
        if maximum_sequence_length is not None:
            return x.shape[1] == maximum_sequence_length
        else:
            return True
    # If the array is 1-dimensional, we need to check if all
    # elements have the same length
    elif x.ndim == 1:
        unique_lengths = set(map(len, x))
        if len(unique_lengths) == 1:
            if maximum_sequence_length is not None:
                return unique_lengths.pop() == maximum_sequence_length
            else:
                return True
        else:
            # Then there's more than one length
            return False
    else:
        return False



from pathlib import Path

from poli.core.problem import Problem
from poli.core.proteins.data_packages import RFPFoldXStabilitySupervisedDataPackage
from poli.core.util.seeding import seed_python_numpy_and_torch
from poli.objective_repository.foldx_stability.register import (
    FoldXStabilityBlackBox,
    FoldXStabilityProblemFactory,
)


class RFPFoldXStabilityBlackBox(FoldXStabilityBlackBox):
    def __init__(
        self,
        experiment_id=None,
        tmp_folder=None,
        eager_repair=False,
        verbose=False,
        batch_size=1,
        parallelize=False,
        num_workers=None,
        evaluation_budget=float("inf"),
        force_isolation=False,
    ):
        RFP_FOLDX_ASSETS_DIR = Path(__file__).parent / "assets"

        wildtype_pdb_path = [
            RFP_FOLDX_ASSETS_DIR / "2vad_A_Repair.pdb",
            RFP_FOLDX_ASSETS_DIR / "2vae_D_Repair.pdb",
            RFP_FOLDX_ASSETS_DIR / "3e5v_A_Repair.pdb",
            RFP_FOLDX_ASSETS_DIR / "3ned_A_Repair.pdb",
            RFP_FOLDX_ASSETS_DIR / "5lk4_A_Repair.pdb",
            RFP_FOLDX_ASSETS_DIR / "6aa7_A_Repair.pdb",
        ]
        super().__init__(
            wildtype_pdb_path,
            experiment_id,
            tmp_folder,
            eager_repair,
            verbose,
            batch_size,
            parallelize,
            num_workers,
            evaluation_budget,
            force_isolation,
        )


class RFPFoldXStabilityProblemFactory(FoldXStabilityProblemFactory):
    def create(
        self,
        experiment_id=None,
        tmp_folder=None,
        eager_repair=False,
        verbose=False,
        seed=None,
        batch_size=1,
        parallelize=False,
        num_workers=None,
        evaluation_budget=float("inf"),
        force_isolation=False,
    ):
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        rfp_foldx = RFPFoldXStabilityBlackBox(
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            eager_repair=eager_repair,
            verbose=verbose,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        x0 = rfp_foldx.x0

        data_package = RFPFoldXStabilitySupervisedDataPackage()

        return Problem(
            black_box=rfp_foldx,
            x0=x0,
            data_package=data_package,
        )



"""This module tests whether forcing the isolation of black
box objectives (i) creates the relevant conda environment, and
(ii) writes the isolated function script in the config,
(iii) getting the inner function works as expected."""

import subprocess

from poli.core.util.isolation.instancing import get_inner_function


def test_force_isolation_on_tdc():
    from poli import objective_factory

    problem = objective_factory.create(
        name="deco_hop",
        force_isolation=True,
    )

    # After creating the problem, we check that:
    # (i) creates the relevant conda environment
    assert "poli__tdc" in subprocess.check_output("conda env list".split()).decode()

    # (ii) writes the isolated function script in the config
    from poli.objective_factory import load_config

    config = load_config()
    assert config["tdc__isolated"]

    # (iii) we can call the inner function
    inner_f = get_inner_function(
        isolated_function_name="tdc__isolated",
        class_name="TDCIsolatedLogic",
        module_to_import="poli.core.chemistry.tdc_isolated_function",
        force_isolation=True,
        oracle_name="deco_hop",
        from_smiles=True,
        quiet=True,
    )
    assert (problem.black_box(problem.x0) == inner_f(problem.x0)).all()



"""
Implements an abstract problem.

Problems are standardized ways of running benchmarks. They are used to
create black boxes and to run optimization algorithms.

At its core, a problem is a black box with a known setup. This setup
includes the following information:

- The initial evaluation(s) x0 and y0.
- The fidelity of the black box (e.g., "high" or "low").
- The black box function and its information (e.g. whether it is
  noisy, continuous or discrete, etc.).
- An evaluation budget, which is the maximum number of evaluations
  allowed.
"""

from __future__ import annotations

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.data_package import DataPackage
from poli.core.util.algorithm_observer_wrapper import AlgorithmObserverWrapper
from poli.core.util.default_observer import DefaultObserver


class Problem:
    def __init__(
        self,
        black_box: AbstractBlackBox,
        x0: np.ndarray,
        data_package: DataPackage | None = None,
        strict_validation: bool = True,
    ):
        self.black_box: AbstractBlackBox = black_box
        self.x0: np.ndarray = x0
        self.data_package: DataPackage | None = data_package
        self.black_box_information = black_box.info
        if strict_validation:
            self._validate()
        self.observer: AlgorithmObserverWrapper = AlgorithmObserverWrapper(
            DefaultObserver()
        )
        self.observer_info = None

    def _validate(self):
        if not isinstance(self.black_box, AbstractBlackBox):
            raise ValueError("Black box must be an instance of AbstractBlackBox.")
        if not isinstance(self.x0, np.ndarray):
            raise ValueError("x0 must be a numpy array.")
        # TODO: validate whether self.x0 is of the right shape.

    def is_discrete(self):
        return self.black_box_information.discrete

    def is_deterministic(self):
        return self.black_box_information.deterministic

    def is_continuous(self):
        return not self.is_discrete()

    def set_observer(self, observer: AlgorithmObserverWrapper, observer_info: object):
        self.observer = observer
        self.observer_info = observer_info

    @property
    def info(self):
        return self.black_box.info

    def __str__(self):
        return f"<Problem: {self.black_box.info.get_problem_name()} x0: {self.x0} data_package: {self.data_package}>"

    def __repr__(self):
        return f"<Problem: {self.black_box.info.get_problem_name()} x0: {self.x0} data_package: {self.data_package} discrete={self.is_discrete()} deterministic={self.is_deterministic()}>"



import io
import os
from pathlib import Path

import numpy as np
import pytest

from poli import objective_factory

THIS_DIR = Path(__file__).parent.resolve()


@pytest.fixture(scope="session", autouse=True)
def cleanup_optin_file():
    file_path = (
        THIS_DIR.parent.parent.parent
        / "objective_repository"
        / "rosetta_energy"
        / ".pyrosetta_accept.txt"
    )
    cleaned_pdbs = THIS_DIR.glob("*.clean.pdb")  # created by Rosetta during runtime

    # for individual tests, the opt-in file is required for use...
    if not file_path.exists():
        with open(file_path, "w") as file:
            file.write("accepted")

    yield  # control to the tests

    if file_path.exists():
        os.remove(file_path)
    for file in cleaned_pdbs:  # cleanup created PDBs
        os.remove(file)


@pytest.mark.poli__rosetta_energy
def test_rosetta_optin(monkeypatch):
    monkeypatch.setattr("sys.stdin", io.StringIO("Y"))
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        relax=False,  # fast compute
        pack=False,
    )
    assert problem is not None


@pytest.mark.poli__rosetta_energy
@pytest.mark.parametrize("unit", ["DDG", "REU", "DREU"])
def test_rosetta_wt_zero_ddg(unit):
    """
    Test that the WT score is equal to x0 evaluated on f
    """
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        relax=False,  # fast compute
        pack=False,
        unit=unit,
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)
    if unit == "REU":
        assert f.inner_function.wt_score == y0
    else:
        assert np.isclose(y0, 0.0)


@pytest.mark.poli__rosetta_energy
def test_rosetta_on_3ned_sequence_mutations_correct():
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        relax=False,  # fast compute
        pack=False,
        clean=False,  # keep sequences as loaded for consistent testing against PDB
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        wildtype_sequence[:4] + "R" + wildtype_sequence[5:],
        wildtype_sequence[:9] + "N" + wildtype_sequence[10:],
    ]

    x = np.array([list(mutation) for mutation in three_mutations])
    f(x)  # function call required to populate x_t property

    # Asserting that the mutations are according to expectations
    # E1A
    # M5R
    # E10N

    for i, mutant in enumerate(three_mutations):
        assert mutant[:20] == f.inner_function.x_t[i][:20]


@pytest.mark.poli__rosetta_energy
def test_rosetta_on_3ned_against_known_results():
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        relax=False,  # fast compute
        pack=False,
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
        "N" + wildtype_sequence[1:],
    ]

    x = np.array([list(mutation) for mutation in three_mutations])
    y = f(x)

    # Asserting that the results are the consistent:
    # E1A: -3.85851953
    # E1R: -3.6878339
    # E1N: -3.80845599

    assert np.isclose(y[0], -3.85851953, atol=1e-4)
    assert np.isclose(y[1], -3.6878339, atol=1e-4)
    assert np.isclose(y[2], -3.80845599, atol=1e-4)


@pytest.mark.poli__rosetta_energy
def test_rosetta_on_3ned_cartesian_against_known_results():
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        score_function="ref2015_cart",
        relax=False,  # fast compute
        pack=False,
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
        "N" + wildtype_sequence[1:],
    ]

    x = np.array([list(mutation) for mutation in three_mutations])
    y = f(x)

    # Asserting that the results are the consistent:
    # E1A: -4.06154293
    # E1R: -3.95461636
    # E1N: -4.08711725

    assert np.isclose(y[0], -4.06154293, atol=1e-4)
    assert np.isclose(y[1], -3.95461636, atol=1e-4)
    assert np.isclose(y[2], -4.08711725, atol=1e-4)


@pytest.mark.poli__rosetta_energy
def test_rosetta_on_3ned_evaluating_twice_vs_once():
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        relax=False,  # fast compute
        pack=False,
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
        "N" + wildtype_sequence[1:],
    ]

    two_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
    ]
    one_mutation = [
        "N" + wildtype_sequence[1:],
    ]

    x_all = np.array([list(mutation) for mutation in three_mutations])
    y = f(x_all)

    y_two = f(np.array([list(mutation) for mutation in two_mutations]))
    y_one = f(np.array([list(mutation) for mutation in one_mutation]))

    assert y[0] == y_two[0]
    assert y[1] == y_two[1]
    assert y[2] == y_one[0]
    # Asserting that the results are the same as in the
    # notebook:
    # E1A: -3.85851953
    # E1R: -3.6878339
    # E1N: -3.80845599

    assert np.isclose(y[0], -3.85851953, atol=1e-4)
    assert np.isclose(y[1], -3.6878339, atol=1e-4)
    assert np.isclose(y[2], -3.80845599, atol=1e-4)


@pytest.mark.poli__rosetta_energy
def test_rosetta_on_3ned_against_results_isolated():
    """
    We test forceful registration of the Rosetta problem.
    """
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        force_isolation=True,
        relax=False,  # fast compute
        pack=False,
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
        "N" + wildtype_sequence[1:],
    ]

    x = np.array([list(mutation) for mutation in three_mutations])
    y = f(x)

    # Asserting that the results are the same as in the
    # notebook:
    # E1A: -3.85851953
    # E1R: -3.6878339
    # E1N: -3.80845599

    assert np.isclose(y[0], -3.85851953, atol=1e-4)
    assert np.isclose(y[1], -3.6878339, atol=1e-4)
    assert np.isclose(y[2], -3.80845599, atol=1e-4)


@pytest.mark.poli__rosetta_energy
@pytest.mark.slow
def test_rosetta_relax_pack():
    problem = objective_factory.create(
        name="rosetta_energy",
        wildtype_pdb_path=THIS_DIR / "3ned.pdb",
        relax=True,  # extended protocol compute
        pack=True,
        seed=0,
    )
    f, x0 = problem.black_box, problem.x0

    wildtype_sequence = "".join(x0[0])
    three_mutations = [
        "A" + wildtype_sequence[1:],
        "R" + wildtype_sequence[1:],
    ]

    x = np.array([list(mutation) for mutation in three_mutations])
    y = f(x)

    # Asserting that the results are the consistent:
    # E1A: 1.87951932
    # E1R: 2.79688494

    assert np.isclose(y[0], 1.87951932, atol=1e-4)
    assert np.isclose(y[1], 2.79688494, atol=1e-4)
